#!/usr/bin/env python
#
#    denss_pdb2mrc.py
#    A tool for calculating simple electron density maps from pdb files.
#
#    Part of the DENSS package
#    DENSS: DENsity from Solution Scattering
#    A tool for calculating an electron density map from solution scattering data
#
#    Tested using Anaconda / Python 2.7
#
#    Author: Thomas D. Grant
#    Email:  <tdgrant@buffalo.edu>
#    Copyright 2017-Present The Research Foundation for SUNY
#
#    Additional Authors:
#    Sarah Chamberlain
#    Stephen Moore
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

import time


import denss

import numpy as np
from scipy import ndimage
import sys, argparse, os, logging
import copy
import time
from textwrap import wrap

def main():
    t = []
    fs = []
    t.append(time.time())
    fs.append("start")

    parser = argparse.ArgumentParser(description="A tool for calculating simple electron density maps from pdb files.", formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("--version", action="version",version="%(prog)s v{version}".format(version=denss.__version__))
    parser.add_argument("-f", "--file", type=str, help="Atomic model as a .pdb file for input (required).")
    parser.add_argument("-d", "--data", type=str, help="Experimental SAXS data file for input (3-column ASCII text file (q, I, err), optional).")
    parser.add_argument("--fast", dest="fast", action="store_true", help="Fast mode. Sets nsamples to 64 (increases voxel size), disables plotting, disables mrc writing, sets shell type to uniform.")
    parser.add_argument("-n1", "--n1", default=None, type=int, help="First data point to use of experimental data")
    parser.add_argument("-n2", "--n2", default=None, type=int, help="Last data point to use of experimental data")
    parser.add_argument("-u", "--units", default="a", type=str, help="Angular units of experimental data (\"a\" [1/angstrom] or \"nm\" [1/nanometer]; default=\"a\"). If nm, will convert output to angstroms.")
    parser.add_argument("-qmin", "--qmin", default=None, type=float, help="Minimum q value to use for fitting experimental data.")
    parser.add_argument("-qmax", "--qmax", default=None, type=float, help="Maximum q value to use for fitting experimental data.")
    parser.add_argument("-nq", "--nq", default=None, type=int, help="Number of q values to include in final scattering profile (only relevant for prediction with no experimental data, uses interpolation).")
    parser.add_argument("-qfile", "--qfile", default=None, type=str, help="Filename of ASCII text file containing desired q values to write as the first column (only relevant for prediction with no experimental data, uses interpolation).")
    parser.add_argument("-s", "--side", default=None, type=float, help="Desired side length of real space box (default=3*Dmax).")
    parser.add_argument("-v", "--voxel", default=None, type=float, help="Desired voxel size (default=1.0)")
    parser.add_argument("-n", "--nsamples", default=None, type=int, help="Desired number of samples (i.e. voxels) per axis (default=variable)")
    parser.add_argument("-b", "--use_b", dest="use_b", action="store_true", help="Include B-factors from atomic model (optional, default=False)")
    parser.add_argument("-B", "--global_B", default=None, type=float, help="Desired global B-factor (added to any individual B-factors if enabled))")
    parser.add_argument("-r", "--resolution", default=None, type=float, help="Desired resolution (after scattering profile calculation, blur map by this width using gaussian kernel, similar to Chimera Volume Filter function).")
    parser.add_argument("-exH","--explicitH", dest="explicitH", action="store_true", help="Use hydrogens in pdb file (optional, default=True if H exists)")
    parser.add_argument("-imH","--implicitH", dest="explicitH", action="store_false", help=argparse.SUPPRESS) #help="Use implicit hydrogens approximation (optional, EXPERIMENTAL)")
    parser.add_argument("-recalc","--recalc","--recalculate_volumes", dest="recalculate_atomic_volumes", action="store_true", help="Calculate atomic volumes directly from coordinates rather than using lookup table (default=False)")
    parser.add_argument("--read_radii", default=False, action="store_true", help="Read adjusted per atom radii (for volume calculation) from Occupancy column of PDB file (default=False)")
    parser.add_argument("-c_on", "--center_on", dest="center", action="store_true", help="Center PDB (default).")
    parser.add_argument("-c_off", "--center_off", dest="center", action="store_false", help="Do not center PDB.")
    parser.add_argument("-iw_on", "--iw_on", "--ignore_waters", "--ignore_waters_on", dest="ignore_waters", action="store_true", help="Ignore waters (default=True).")
    parser.add_argument("-iw_off", "--iw_off", "--ignore_waters_off", dest="ignore_waters", action="store_false", help="Turn Ignore waters off (i.e., read the waters).")
    parser.add_argument("-rho0", "--rho0", default=0.334, type=float, help="Density of bulk solvent in e-/A^3 (default=0.334)")
    parser.add_argument("-exvol_type", "--exvol_type", default="gaussian", type=str, help="Type of excluded volume (gaussian (default) or flat)")
    parser.add_argument("-fit_off", "--fit_off", dest="fit_all", action="store_false", help="Do not fit either rho0 or shell contrast (optional)")
    parser.add_argument("-fit_rho0_on", "--fit_rho0_on", dest="fit_rho0", action="store_true", help="Fit rho0, the bulk solvent density (optional, default=True)")
    parser.add_argument("-fit_rho0_off", "--fit_rho0_off", dest="fit_rho0", action="store_false", help="Do not fit rho0, the bulk solvent density (optional, default=True)")
    parser.add_argument("-fit_shell_on", "--fit_shell_on", dest="fit_shell", action="store_true", help="Fit hydration shell parameters (optional, default=True)")
    parser.add_argument("-fit_shell_off", "--fit_shell_off", dest="fit_shell", action="store_false", help="Do not fit hydration shell parameters (optional, default=True)")
    parser.add_argument("-drho","--drho","-shell","-shell_contrast", "--shell_contrast", dest="shell_contrast", default=0.011, type=float, help="Initial mean contrast of hydration shell in e-/A^3 (default=0.011)")
    parser.add_argument("-shell_type", "--shell_type", default=None, type=str, help="Type of hydration shell (water (default) or uniform)")
    parser.add_argument("-shell_mrcfile", "--shell_mrcfile", default=None, type=str, help=argparse.SUPPRESS) #help="Filename of hydration shell mrc file (default=None)")
    parser.add_argument("-fit_radii", "--fit_radii", dest="fit_radii", action="store_true", help=argparse.SUPPRESS) #help="Fit atomic radii for excluded volume calculation (optional, default=False)")
    parser.add_argument("--radii_sf", default=None, type=float, nargs='+', help=argparse.SUPPRESS) #help="Atomic radii scale factors for excluded volume calculation (optional, ordered as [H C N O])")
    parser.add_argument("--set_radii_explicitly", default=None, type=str, help=argparse.SUPPRESS) #help="Set atomic radii explicitly for different atom types. Ignores --radii_sf. Format H:1.07:C:1.58:N:0.084:O:1.30"
    parser.add_argument("-fit_scale_on", "--fit_scale_on", dest="fit_scale", action="store_true", help="Include scale factor in least squares fit to data (optional, default=True)")
    parser.add_argument("-fit_scale_off", "--fit_scale_off", dest="fit_scale", action="store_false", help="Do not include offset in least squares fit to data.")
    parser.add_argument("-fit_offset_on", "--fit_offset_on", dest="fit_offset", action="store_true", help="Include offset in least squares fit to data (optional, default=False)")
    parser.add_argument("-fit_offset_off", "--fit_offset_off", dest="fit_offset", action="store_false", help="Do not include offset in least squares fit to data.")
    parser.add_argument("-p", "-penalty_weight", "--penalty_weight", default=None, type=float, help="Overall penalty weight for fitting parameters (default=0)")
    parser.add_argument("-ps", "-penalty_weights", "--penalty_weights", default=[1.0, 0.01], type=float, nargs='+', help="Individual penalty weights for each parameter (space separated listed of weights for [rho0,shell], default=1.0 0.01)")
    parser.add_argument("-min_method", "--min_method", "-minimization_method","--minimization_method", dest="method", default='Nelder-Mead', type=str, help="Minimization method (scipy.optimize method, default=Nelder-Mead).")
    parser.add_argument("-min_options", "--min_options", "-minimization_options","--minimization_options", dest="minopts", default='{"adaptive": True}', type=str, help="Minimization options (scipy.optimize options formatted as python dictionary, default=\"{'adaptive': True}\").")
    parser.add_argument("-write_on", "--write_on", action="store_true", dest="write_mrc_file", help="Write MRC file (default=True).")
    parser.add_argument("-write_off", "--write_off", action="store_false", dest="write_mrc_file", help="Do not write MRC file.")
    parser.add_argument("-write_extras", "--write_extras", action="store_true", default=False, help="Write out extra MRC files for invacuo, exvol, shell densities (default=False).")
    parser.add_argument("-write_pdb_on", "--write_pdb_on", action="store_true", dest="write_pdb", help="Write modified pdb file including recentering and atom radii to B-factor column (default=True).")
    parser.add_argument("-write_pdb_off", "--write_pdb_off", action="store_false", dest="write_pdb", help="Do not write modified pdb file.")
    parser.add_argument("-interp_on", "--interp_on", dest="Icalc_interpolation", action="store_true", help="Interpolate I_calc to experimental q grid (default).")
    parser.add_argument("-interp_off", "--interp_off", dest="Icalc_interpolation", action="store_false", help="Do not interpolate I_calc to experimental q grid .")
    parser.add_argument("--use_sasrec_during_fitting", default=False, action="store_true", help="Use Sasrec during fitting procedure for interpolation (more accurate, slower. Note: Sasrec is always used for the output .dat or .fit file, this is just for during the iterative fitting procedure.).")
    parser.add_argument("--plot_on", dest="plot", action="store_true", help="Create simple plots of results (requires Matplotlib, default if module exists).")
    parser.add_argument("--plot_off", dest="plot", action="store_false", help="Do not create simple plots of results. (Default if Matplotlib does not exist)")
    parser.add_argument("--print_timings", default=False, action="store_true", help="Print timings for each step of the script.")
    parser.add_argument("-o", "--output", default=None, help="Output filename prefix (default=basename_pdb)")
    parser.add_argument("--write_shannon", dest="write_shannon", action="store_true", help=argparse.SUPPRESS) # help="Write a file containing only the Shannon intensities.")
    parser.set_defaults(fast = False)
    parser.set_defaults(ignore_waters = True)
    parser.set_defaults(center = True)
    parser.set_defaults(plot=None)
    parser.set_defaults(use_b=False)
    parser.set_defaults(explicitH=None)
    parser.set_defaults(recalculate_atomic_volumes=False)
    parser.set_defaults(fit_rho0=True)
    parser.set_defaults(fit_shell=True)
    parser.set_defaults(fit_all=True)
    parser.set_defaults(fit_radii=False)
    parser.set_defaults(fit_scale=True)
    parser.set_defaults(fit_offset=False)
    parser.set_defaults(Icalc_interpolation=True)
    parser.set_defaults(write_mrc_file=None)
    parser.set_defaults(write_pdb=None)
    parser.set_defaults(write_shannon=False)
    args = parser.parse_args()

    np.set_printoptions(linewidth=150,precision=10)

    t.append(time.time())
    fs.append("argparse")

    start = time.time()

    scriptname = os.path.basename(sys.argv[0])
    command = scriptname + ' ' + ' '.join(sys.argv[1:])

    fname_nopath = os.path.basename(args.file)
    basename, ext = os.path.splitext(fname_nopath)

    if args.output is None:
        output = basename + "_pdb"
    else:
        output = args.output

    logging.basicConfig(filename=output+'.log',level=logging.INFO,filemode='w',
                        format='%(asctime)s %(message)s') #, datefmt='%Y-%m-%d %I:%M:%S %p')
    logging.info('BEGIN')
    logging.info('Command: %s', ' '.join(sys.argv))
    logging.info('DENSS Version: %s', denss.__version__)
    # logging.info('PDB filename: %s', args.file)

    t.append(time.time())
    fs.append("init log")

    pdb = denss.PDB(args.file, ignore_waters=args.ignore_waters)

    t.append(time.time())
    fs.append("read pdb")

    if args.read_radii:
        #if the file has unique radius in occupancy column, use it
        pdb.unique_radius = pdb.occupancy
        pdb.unique_volume = 4/3*np.pi*pdb.unique_radius**3

    #if pdb contains an H atomtype, set explicit hydrogens to True
    if args.explicitH is None and np.any(np.core.defchararray.find(pdb.atomtype,"H")!=-1):
        #if explicitH not set, and there are hydrogens, set explicitH to true
        args.explicitH = True
    elif args.explicitH is None:
        #if explicitH not set, and there are no hydrogens, set explicitH to false, to use implicit hydrogens
        args.explicitH = False
        print("#"*90)
        print("#  WARNING: use of implicit hydrogens is an experimental feature.                        #")
        print("#  Recommend adding explicit hydrogens using a tool such as Reduce, CHARMM, PyMOL, etc.  #")
        print("#"*90)

    if args.fast:
        # only set these values in fast mode if they aren't explicitly set in the command line
        if args.nsamples is None:
            args.nsamples = 64
        if args.write_mrc_file is None:
            args.write_mrc_file = False
        if args.shell_type is None:
            args.shell_type = "uniform"
        if args.plot is None:
            args.plot = False
        if args.write_pdb is None:
            args.write_pdb = False
    else:
        if args.write_mrc_file is None:
            args.write_mrc_file = True
        if args.plot is None:
            args.plot = True
        if args.write_pdb is None:
            args.write_pdb = True

    if args.plot:
        # if plotting is enabled, try to import matplotlib
        # if import fails, set plotting to false
        try:
            import matplotlib.pyplot as plt
            from matplotlib import gridspec
        except ImportError as e:
            print("matplotlib import failed.")
            args.plot = False

    t.append(time.time())
    fs.append("import matplotlib")

    pdb2mrc = denss.PDB2MRC(
        pdb=pdb,
        ignore_waters=args.ignore_waters,
        explicitH=args.explicitH,
        modifiable_atom_types=None,
        center_coords=args.center,
        radii_sf=args.radii_sf,
        recalculate_atomic_volumes=args.recalculate_atomic_volumes,
        exvol_type=args.exvol_type,
        use_b=args.use_b,
        global_B=args.global_B,
        resolution=args.resolution,
        voxel=args.voxel,
        side=args.side,
        nsamples=args.nsamples,
        rho0=args.rho0,
        shell_contrast=args.shell_contrast,
        shell_mrcfile=args.shell_mrcfile,
        shell_type=args.shell_type,
        Icalc_interpolation=args.Icalc_interpolation,
        fit_scale=args.fit_scale,
        fit_offset=args.fit_offset,
        data_filename=args.data,
        data_units=args.units,
        n1=args.n1,
        n2=args.n2,
        qmin=args.qmin,
        qmax=args.qmax,
        penalty_weight=args.penalty_weight,
        penalty_weights=args.penalty_weights,
        fit_rho0=args.fit_rho0,
        fit_shell=args.fit_shell,
        fit_all=args.fit_all,
        min_method=args.method,
        min_opts=args.minopts,
        fast=args.fast,
        use_sasrec_during_fitting=args.use_sasrec_during_fitting,
        )

    t.append(time.time())
    fs.append("pdb2mrc init")

    #write the modified pdb file, and store the 
    #new unique volume/radius value in the occupancy
    #column, to prevent needing to recalculate if wanting
    #to run the fitting again.
    pdboutput = basename
    if not args.read_radii:
        pdboutput += '_out.pdb'
        pdbout = copy.deepcopy(pdb2mrc.pdb)
        pdbout.occupancy = pdb2mrc.pdb.unique_radius
        if args.write_pdb:
            pdbout.write(filename=pdboutput)

    t.append(time.time())
    fs.append("log stuff")

    pdb2mrc.scale_radii(radii_sf=args.radii_sf)
    t.append(time.time())
    fs.append("scale_radii")

    if args.set_radii_explicitly is not None:
        arg_list = args.set_radii_explicitly.split(":")
        atom_types = arg_list[::2]
        radii = np.array(arg_list[1::2], dtype=float)
        pdb2mrc.set_radii(atom_types, radii)
    pdb2mrc.make_grids()
    t.append(time.time())
    fs.append("make_grids")

    pdb2mrc.calculate_global_B()
    t.append(time.time())
    fs.append("calculate_global_B")

    print("Optimal Side length >= %.2f" % pdb2mrc.optimal_side)
    print("Optimal N samples   >= %d" % pdb2mrc.optimal_nsamples)
    print("Optimal Voxel size  <= %.4f" % pdb2mrc.optimal_voxel)
    print("Actual  Side length  = %.2f" % pdb2mrc.side)
    print("Actual  N samples    = %d" % pdb2mrc.n)
    print("Actual  Voxel size   = %.4f" % pdb2mrc.dx)
    print("Global B-factor      = %.4f" % pdb2mrc.global_B)

    pdb2mrc.calculate_invacuo_density()
    t.append(time.time())
    fs.append("calculate_invacuo_density")

    pdb2mrc.calculate_excluded_volume()
    t.append(time.time())
    fs.append("calculate_excluded_volume")

    pdb2mrc.calculate_hydration_shell()
    t.append(time.time())
    fs.append("calculate_hydration_shell")

    pdb2mrc.calculate_structure_factors()
    t.append(time.time())
    fs.append("calculate_structure_factors")

    if args.data is not None:
        pdb2mrc.load_data()
        t.append(time.time())
        fs.append("load_data")
        pdb2mrc.initialize_penalties(penalty_weight=args.penalty_weight)
        pdb2mrc.minimize_parameters(fit_radii=args.fit_radii)
    t.append(time.time())
    fs.append("minimize_parameters")

    print()
    print("Final parameter values:")
    for i in range(len(pdb2mrc.params)):
        print("%s : %.5e" % (pdb2mrc.param_names[i],pdb2mrc.params[i]))

    pdb2mrc.calc_rho_with_modified_params(pdb2mrc.params)
    pdb2mrc.shell_mean_density = np.mean(pdb2mrc.rho_shell[pdb2mrc.water_shell_idx])
    t.append(time.time())
    fs.append("calc_rho_with_modified_params")

    pdb2mrc.calc_I_with_modified_params(pdb2mrc.params)
    t.append(time.time())
    fs.append("calc_I_with_modified_params")
    if args.data:
        pdb2mrc.calc_chi2()
        fit = pdb2mrc.fit
        optimized_chi2 = pdb2mrc.optimized_chi2
        exp_scale_factor = pdb2mrc.exp_scale_factor
        offset = pdb2mrc.offset
        print("Scale factor: %.5e " % pdb2mrc.exp_scale_factor)
        print("Offset: %.5e " % pdb2mrc.offset)
        print("chi2 of fit:  %.5e " % pdb2mrc.optimized_chi2)

    print("Calculated average radii:")
    pdb2mrc.calculate_average_radii()
    t.append(time.time())
    fs.append("calculate_average_radii")
    for i in range(len(pdb2mrc.modifiable_atom_types)):
        print("%s: %.3f"%(pdb2mrc.modifiable_atom_types[i],pdb2mrc.mean_radius[i]))

    pdb2mrc.calculate_excluded_volume_in_A3()
    print("Calculated excluded volume: %.2f"%(pdb2mrc.exvol_in_A3))

    end = time.time()
    print("Total calculation time: %.3f seconds" % (end-start))

    if args.qfile is not None:
        Iq = np.genfromtxt(args.qfile)
        qc = Iq[:,0]
    else:
        qc = None

    # always use sasrec interpolation for final output calculated profile, since its more accurate
    pdb2mrc.save_Iq_calc(prefix=output, qmax=args.qmax, nq=args.nq, qc=qc, use_sasrec=True)

    if args.write_shannon:
        qcalc_max = pdb2mrc.Iq_calc[:,0].max()
        wshannon = np.pi/pdb2mrc.D
        nshannon = int(qcalc_max/wshannon)
        qshannon = np.linspace(0,nshannon*wshannon,nshannon)
        pdb2mrc.save_Iq_calc(prefix=output+'_Shannon', qc=qshannon)

    if args.data is not None:
        pdb2mrc.save_fit(prefix=output)

        if args.plot:
            fig = plt.figure(figsize=(8, 6))
            gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
            ax0 = plt.subplot(gs[0])
            ax1 = plt.subplot(gs[1], sharex=ax0)

            q = fit[:,0]
            Ie = fit[:,1]
            err = fit[:,2]
            Ic = fit[:,3]

            ax0.plot(q, Ie,'.',c='gray',label=pdb2mrc.data_filename)
            ax0.plot(q, Ic, '-',c='red',label=basename+'.fit \n' + r'$\chi^2 = $ %.2f'%optimized_chi2)
            resid = (Ie - Ic)/err
            ax1.plot(q, resid*0, 'k--')
            ax1.plot(q, resid, '.',c='red')

            ax0.semilogy()
            ax1.set_xlabel(r"q ($\AA^{-1}$)")
            ax0.set_ylabel("I(q)")
            ax1.set_ylabel(r"$\Delta I / \sigma$")
            fig.suptitle(output)
            #title is often long, so wrap it to multiple lines if needed
            title = "\n".join(wrap(command, 80))
            ax0.set_title(title)
            ax0.legend()
            plt.tight_layout()
            plt.savefig(output+'_fits.png',dpi=300)
            # plt.show()
            plt.close()
    t.append(time.time())
    fs.append("plotting")

    #after all the map calculations are done, apply any resolution-based blurring directly to the map
    #but this should not be included in the scattering profile calculation, so we do it here just
    #before writing the maps
    if args.resolution is not None:
        sigma = args.resolution / pdb2mrc.dx
        if sigma <= 0.5:
            print("Note: requested resolution is less than half a voxel, meaning nothing will be done for the Gaussian filter.")
        if args.write_mrc_file:
            ne = pdb2mrc.rho_insolvent.sum()
            pdb2mrc.rho_insolvent = ndimage.gaussian_filter(pdb2mrc.rho_insolvent, sigma, mode='wrap')
            pdb2mrc.rho_insolvent *= ne/pdb2mrc.rho_insolvent.sum()
        if args.write_extras:
            ne = pdb2mrc.rho_invacuo.sum()
            pdb2mrc.rho_invacuo = ndimage.gaussian_filter(pdb2mrc.rho_invacuo, sigma, mode='wrap')
            pdb2mrc.rho_invacuo *= ne/pdb2mrc.rho_invacuo.sum()

            ne = pdb2mrc.rho_exvol.sum()
            pdb2mrc.rho_exvol = ndimage.gaussian_filter(pdb2mrc.rho_exvol, sigma, mode='wrap')
            pdb2mrc.rho_exvol *= ne/pdb2mrc.rho_exvol.sum()

            ne = pdb2mrc.rho_shell.sum()
            pdb2mrc.rho_shell = ndimage.gaussian_filter(pdb2mrc.rho_shell, sigma, mode='wrap')
            pdb2mrc.rho_shell *= ne/pdb2mrc.rho_shell.sum()
    t.append(time.time())
    fs.append("gaussian_filter")

    if args.write_mrc_file:
        #write output
        print('Writing density map to %s_insolvent.mrc file.'%output)
        denss.write_mrc(pdb2mrc.rho_insolvent/pdb2mrc.dV,pdb2mrc.side,output+"_insolvent.mrc")
    if args.write_extras:
        denss.write_mrc(pdb2mrc.rho_invacuo/pdb2mrc.dV,pdb2mrc.side,output+"_invacuo.mrc")
        denss.write_mrc(pdb2mrc.rho_exvol/pdb2mrc.dV,pdb2mrc.side,output+"_exvol.mrc")
        denss.write_mrc(pdb2mrc.rho_shell/pdb2mrc.dV,pdb2mrc.side,output+"_shell.mrc")
    t.append(time.time())
    fs.append("write_mrc")

    if args.print_timings:
        print("%40s: %7s %7s"%("function","time","cumtime"))
        for i in range(1,len(t)):
            print("%40s: %7.3f %7.3f"%(fs[i],t[i]-t[i-1],t[i]-t[0]))


if __name__ == "__main__":
    main()





"""
    DataUtils.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF
"""
import sys
import os
import glob
import re

def get_pytools_folder():
    delimter = '\\'
    file_nodes = os.path.abspath( __file__ ).split( delimter )
    try:
        i = file_nodes.index( 'PyTools' ) + 1
    except:
        file_nodes = os.path.dirname( os.path.abspath( __file__ ) ).split( delimter )
        file_nodes[-4] = 'PyTools'
        i = len(file_nodes) - 3

    root_dir = delimter.join( file_nodes[0:i] )
    return root_dir.replace( delimter, '/' )

def cut_upper_folders(folder):
    return re.sub(r'^\w:.*/Data/', '', folder)

def get_in_folder(in_folder=None):
    if in_folder is None:
        from molass_legacy._MOLASS.SerialSettings import get_setting
        in_folder = get_setting('in_folder')
        if in_folder is None:
            return "unkown folder"      # possibly, it is not set yet as in a separate process
    nodes = in_folder.replace("\\", "/").split("/")
    if len(nodes) > 4:
        in_folder = "/".join(["…"] + nodes[-3:])
    return in_folder

def compact_path_name(path, separator='/'):
    nodes = path.split(separator)
    if len(nodes) > 4:
        return separator.join(nodes[0:2] + ['...'] + nodes[-2:])
    else:
        return path

def do_folder( in_folder, cb ):
    dat_files = glob.glob( in_folder + '/*.dat' )
    num_files = len( dat_files )
    if num_files < 90:
        return True

    txt_files = glob.glob( in_folder + '/*.txt' ) + glob.glob( in_folder + '/*.mtd' )
    num_txt_files = len( txt_files )
    if num_txt_files == 0:
        uv_folder = '/'.join( in_folder.split( '/' )[0:-1] ) + '/spectra'
        if os.path.exists( uv_folder ):
            # print( uv_folder )
            pass
        else:
            allow = False
            for name in ['20160628', '20191031', '_no_UV']:
                if in_folder.find(name) >= 0:
                    uv_folder = in_folder
                    allow = True
                    break
            if allow:
                pass
            else:
                print('--------------------- skipping', in_folder)
                return True
    else:
        uv_folder = in_folder

    # print( in_folder, num_files, num_txt_files, uv_folder )
    applied, sd = cb( in_folder, uv_folder=uv_folder, plot=False )
    return applied

def serial_folder_walk(folder, cb, level=0, depth=3, reverse=False):
    for node in sorted( os.listdir( folder ), reverse=reverse ):
        path = '/'.join( [ folder, node ] )
        if not os.path.isdir( path ): continue
        # print(node)

        if node.find( 'autorg' ) >= 0: continue
        if node.find( 'analysis' ) >= 0: continue
        if node.find( 'averaged' ) >= 0: continue
        if node.find( '-stain' ) >= 0: continue
        if node.find( '_original' ) >= 0: continue
        if node.find( 'CirAve' ) >= 0: continue
        if node.find( '00_model' ) >= 0: continue
        if node.find( '_full' ) >= 0: continue
        if node.find( '-skip' ) >= 0: continue

        # print('path=', path)
        if path.find( '20170209_2' ) > 0: continue
        if path.find( '/JT' ) >= 0: continue
        if path.find( '20211105' ) > 0: continue

        ret = do_folder( path, cb )
        if not ret:
            return False
        if level < depth:
            ret = serial_folder_walk( path, cb, level=level+1, depth=depth, reverse=reverse )
            if not ret:
                return False

        if node.find( 'sample_data_no_UV' ) >= 0:
            print("avoid continuing beyond this folder")
            return False
    return True

def mct_folder_walk(folder, cb, level=0, depth=3):
    if not os.path.exists(folder):
        print("%s doed not exist." % folder)
        return True

    for node in sorted( os.listdir( folder ) ):
        if level == 0 and node.find( '20190522' ) < 0: continue
        if level > 0 and node.find( 'Individual' ) < 0: continue

        path = '/'.join( [ folder, node ] )
        if not os.path.isdir( path ): continue
        ret = do_folder( path, cb )
        if not ret:
            return False

        if level < depth:
            ret = serial_folder_walk( path, cb, level=level+1, depth=depth )
            if not ret:
                return False

    return True

def is_microfluidic(folder):
    if folder.find("microfluidic") >= 0:
        return True

    mtd_files = glob.glob(folder + r"\*.mtd")
    return len(mtd_files) > 0

def get_next_subfolder( folder, search_str ):

    found = [ False ]
    next_folder = []
    def func( in_folder, uv_folder, plot ):
        # print( in_folder )
        if found[0]:
            next_folder.append( in_folder )
            return False, None
        else:
            if in_folder.find( search_str ) >= 0:
                found[0] = True
            return True, None

    serial_folder_walk( folder, func )

    if len(next_folder) > 0:
        return next_folder[0]
    else:
        return None

def get_root_folder():
    from molass_legacy.KekLib.ImportUtils import import_module_from_path
    this_dir = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(this_dir + '/../../..', 'TestEnv.py')
    module = import_module_from_path('TestEnv', path)
    env_dict = module.env_dict
    return env_dict['root']

def test_serial_folder_walk():
    roor_folder = get_root_folder()
    def func( in_folder, uv_folder, plot ):
        print( in_folder )
        return True, None

    serial_folder_walk( roor_folder, func )

def test_get_next_subfolder():
    roor_folder = get_root_folder()
    print( get_next_subfolder( roor_folder, 'Sugiyama' ) )

def get_local_path(node_name):
    import platform
    node_dict = {
        'PFSAXS10' : 'D:\\',
        'DESKTOP-DTT9QDU' : 'C:\\Users\\takahashi\\',
        'MasatsuyoPC' : 'E:\\',
    }
    hostname = platform.uname()[1]
    return node_dict[hostname] + node_name

if __name__ == '__main__':
    sys.path.append( os.path.join(os.path.dirname(__file__), '..') )
    # test_serial_folder_walk()
    # test_get_next_subfolder()
    print(get_local_path('TODO'))


"""

    OptimalSmoothing.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF

"""
import numpy as np
from matplotlib.gridspec import GridSpec
from molass_legacy.KekLib.SciPyCookbook import smooth
from scipy.interpolate import UnivariateSpline, LSQUnivariateSpline
from scipy.stats import linregress
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Elution.CurveUtils import MAJOR_PEAK_RATIO
from molass_legacy.KekLib.ExceptionTracebacker import log_exception

SMOOTHING_WINDOW_LEN    = 10
MODEST_NOISE_LEVEL_LIM  = 0.01
NUM_POINTS_PER_KNOT     = 10
MIN_NUM_KNOTS           = 50    # > 40 for pH6
MAX_NUM_KNOTS           = 100
NUM_KONTS_LOW_QUALITY   = 20
MIN_NEIGHBOR_DIFF_RATIO = 0.02
USE_REFINED_BEFORE_CHK  = True
MICROFLUIDIC_START_POS  = 0.5
CORRECT_FOR_FEATURE     = False
CORRECT_LIMIT_RATIO     = 0.05
NEGATIVE_D2_LIMIT_RATIO = 0.01      # 
REFINE_DELTA_X_RATIO    = 0.05
DELTA_X_LIMIT_RATIO     = 0.025     # < 0.026  for SUB_TRN1, < 0.0285 for 20180206
DELTA_Y_LIMIT_RATIO     = 0.01      # < 0.0011 for SUB_TRN1 > 0.005 for 20190607_1, > 0.01 for 20170304
SMALL_DIFF = 1e-6

"""
SUB_TRN1 ratios

original
    [4] 80.87661531533368 (0.01135730305678646, 0.06339442040371927) True

analysis copy
    [4] 72.0857775324413 (3.510140665787276e-05, 0.006980129693787952) False
    [6] 81.22860618043339 (0.0011093295770500207, 0.0265100851340219) True

"""

WINDOW_LEN_PEAK_RECOG_MIN = 20
WINDOW_LEN_PEAK_RECOG_MAX = 40
MAX_DENOM_DELTA_R_RATIO = 500

class OptimalSmoothing:
    def __init__(self, x, y, height=None, bottom=None, min_y_ratio=None, orig_top_x=None):
        self.logger = None
        self.x  = x
        self.y  = y
        self.orig_top_x = orig_top_x
        if height is None:
            height = np.max(y) - np.min(y)
        self.height = height

        if bottom is None:
            bottom = np.min(y)

        self.bottom = bottom
        self.delta_x = DELTA_X_LIMIT_RATIO * len(self.x)

        if min_y_ratio is None:
            min_y_ratio = DELTA_Y_LIMIT_RATIO

        self.min_y_ratio = min_y_ratio
        self.delta_y =  min_y_ratio * height

        min_num_knots = len(x)//2 if len(x) < 100 else MIN_NUM_KNOTS
        num_knots = min( MAX_NUM_KNOTS, max( min_num_knots, len(x)//NUM_POINTS_PER_KNOT ) )
        # num_knots = NUM_KONTS_LOW_QUALITY
        self.knots = np.linspace(x[0], x[-1], num_knots+2)

        if False:
            spline = UnivariateSpline(x, y, s=0)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("OptimalSmoothing.__init__ knots debug")
                ax.plot(x, y)
                ax.plot(self.knots, spline(self.knots), "o")
                fig.tight_layout()
                plt.show()

    def get_logger(self):
        if self.logger is None:
            import logging
            self.logger = logging.getLogger(__name__)

    def plot_variations(self):
        import molass_legacy.KekLib.DebugPlot as plt
        from DataUtils import get_in_folder
        x = self.x
        y = self.y
        in_folder = get_in_folder()

        fig = plt.figure(figsize=(20, 10))
        fig.suptitle("Variations of smoothing window for " + in_folder, fontsize=20)

        gs = GridSpec(3, 4)

        wlen_table = np.array([
            [5,  10, 15, 20],
            [25, 30, 35, 40],
            [50, 70, 100, 200],
            ])

        for i in range(wlen_table.shape[0]):
            for j in range(wlen_table.shape[1]):
                ax = fig.add_subplot(gs[i,j])
                axt = ax.twinx()
                window_len = wlen_table[i, j]

                if window_len > len(y):
                    continue

                ax.set_title("window_len=%d" % window_len)
                ax.plot(x, y)
                sm_y = smooth(y, window_len=window_len)
                ax.plot(x, sm_y)
                spline = UnivariateSpline(x, sm_y, s=0, ext=3)
                d1 = spline.derivative(1)
                d1_y = d1(x)
                axt.plot(x, d1_y, ':', color='green')
                d1_spline = UnivariateSpline(x, d1_y, s=0, ext=3)
                d1_roots = d1_spline.roots()
                d2 = spline.derivative(2)

                peak_x = d1_roots[ np.logical_and( spline(d1_roots) - self.bottom > self.height*0.05, d2(d1_roots) < 0 ) ]
                ax.plot(peak_x, spline(peak_x), 'o', color='red', label='d1_roots')

                feature_spline = LSQUnivariateSpline( x, sm_y, self.knots[1:-1], ext=3 )
                # f_d2 = feature_spline.derivative(2)
                f_d2 = d2
                d2_y = f_d2(x)
                d2_spline = UnivariateSpline(x, d2_y, s=0, ext=3)
                d2_roots = d2_spline.roots()
                d1_max = np.max( np.abs(d1_y) )
                feature_x = d2_roots[ np.logical_and( spline(d2_roots) > self.height*0.03, np.abs(d1(d2_roots))/d1_max > 0.1) ]

                ax.plot(feature_x, spline(feature_x), 'o', color='yellow', label='d2_roots')
                ax.legend()

        fig.tight_layout()
        fig.subplots_adjust( top=0.92 )
        plt.show()

    def compute_optimal_curves(self, microfluidic=False, debug=False):
        x = self.x
        y = self.y

        approximate_peak_num_points = len(y)//8
        num_points = approximate_peak_num_points//4 + 2
        window_len = min(SMOOTHING_WINDOW_LEN, num_points)

        y_for_interpolate = smooth(y, window_len=window_len)
        spline = UnivariateSpline(x, y_for_interpolate, s=0, ext=3)

        noise_level = np.std(y - y_for_interpolate) / self.height

        self.noise_level = noise_level
        self.spline_y = y_for_interpolate
        self.spline = spline

        window_len = min(WINDOW_LEN_PEAK_RECOG_MAX, max(WINDOW_LEN_PEAK_RECOG_MIN, approximate_peak_num_points//2))
        # print('noise_level=', noise_level, 'window_len=', window_len)
        average_left_y = np.average(y[0:approximate_peak_num_points])
        average_right_y = np.average(y[-approximate_peak_num_points:])
        end_diff_ratio = abs(average_left_y - average_right_y)/self.height

        if CORRECT_FOR_FEATURE:
            if end_diff_ratio > CORRECT_LIMIT_RATIO:
                to_be_smoothed_y = self.baseline_correct_simply(x, y)
            else:
                to_be_smoothed_y = y
        else:
            to_be_smoothed_y = y

        y_for_peak_recog = smooth(to_be_smoothed_y, window_len=window_len)
        try:
            spline_peak_recog = LSQUnivariateSpline( x, y_for_peak_recog, self.knots[1:-1], ext=3 )
        except Exception as exc:
            self.get_logger()
            log_exception(self.logger, "LSQUnivariateSpline failed, knots=%s : " % str(self.knots) )
            if False:
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("compute_optimal_curves LSQUnivariateSpline debug")
                    ax.plot(x, y_for_peak_recog)
                    ax.plot(self.knots[1:-1], spline(self.knots[1:-1]), "o")
                    fig.tight_layout()
                    plt.show()
            raise exc
        # spline_peak_recog = UnivariateSpline(x, y_for_peak_recog, s=0, ext=3)

        self.feature_y = y_for_peak_recog
        self.spline_peak_recog = spline_peak_recog

        d1 = spline_peak_recog.derivative(1)
        d1y = d1(x)
        d1_spline = UnivariateSpline(x, d1y, s=0, ext=3)
        self.d1 = d1_spline
        self.d1y = d1y
        self.d1_roots = d1_spline.roots()

        d2 = spline_peak_recog.derivative(2)
        d2y = d2(x)
        d2_spline = UnivariateSpline(x, d2y, s=0, ext=3)
        self.d2 = d2_spline
        self.d2y = d2y
        self.d2_roots = d2_spline.roots()

        if microfluidic:
            start_pos = int(MICROFLUIDIC_START_POS * x[-1])
            n = np.argmax(y_for_interpolate[start_pos:])
            self.top_x = np.array([start_pos + n])
            return

        # print('orig_top_x=', self.orig_top_x)
        # debug = self.orig_top_x is not None
        try:
            top_cadidates = self.get_good_top_candidates(debug=debug)
        except:
            import inspect
            for frm in inspect.stack()[1:]:
                print("---- : %s %s (%d)" % (frm.filename, frm.function, frm.lineno))
            self.get_logger()
            log_exception(self.logger, "get_good_top_candidates failed: ", n=10)
            top_cadidates = []
        if len(top_cadidates) == 0:
            top_cadidates = self.get_original_top_x()
        # print('top_cadidates=', top_cadidates)
        # assert len(top_cadidates) > 0

        sd1 = spline.derivative(1)
        sd1y = sd1(x)
        sd1_spline = UnivariateSpline(x, sd1y, s=0, ext=3)
        sd1_roots = sd1_spline.roots()
        refine_delta_x = REFINE_DELTA_X_RATIO * len(self.x)

        def get_refined_top_points_impl(tobe_refined):
            pk = 5
            refined = []
            last_refined = None
            for p in tobe_refined:
                dist = np.abs(sd1_roots - p)
                neighbor_roots_ = sd1_roots[dist < refine_delta_x]
                pk = min(5, len(neighbor_roots_)-1)
                kk = np.argpartition(np.abs(neighbor_roots_ - p), pk)[0:pk+1]
                if len(kk) > 0:
                    n = np.argmax( self.spline(neighbor_roots_[kk]) )
                    refined_x = neighbor_roots_[kk[n]]
                else:
                    # is this right?
                    refined_x = p
                # print([n], refined_x)
                if last_refined is None or abs(refined_x - last_refined) > SMALL_DIFF:
                    # this check is required for 20180617,
                    # where [4] 158.38825643054835 occurred twice consecutively
                    refined.append(refined_x)
                    last_refined = refined_x

                # if abs(p - refined_x) > 5:
                if debug:
                    import molass_legacy.KekLib.DebugPlot as plt
                    print([p], tobe_refined)
                    x = self.x
                    y = self.y
                    sy = self.spline_y
                    with plt.Dp():
                        fig, ax = plt.subplots()
                        ax.set_title("get_refined_top_points debug")
                        ax.plot(x, y, label='data')
                        ax.plot(x, sy, label='spline')
                        ax.plot(neighbor_roots_, self.spline(neighbor_roots_), 'o', label='neighbor_roots_' )
                        kk_x = neighbor_roots_[kk]
                        ax.plot(kk_x, self.spline(kk_x), 'o', label='neighbor_roots_[kk]' )
                        ax.plot(p, self.spline(p), 'o', color='yellow', label='p' )
                        ax.plot(refined_x, self.spline(refined_x), 'o', color='red', label='refined_x' )
                        ax.legend()
                        plt.show()

            return np.array(refined)

        def get_refined_top_points(tobe_refined):
            try:
                return get_refined_top_points_impl(tobe_refined)
            except:
                # ValueError: attempt to get argmax of an empty sequence at
                # n = np.argmax( self.spline(neighbor_roots_[kk]) )
                log_exception(self.logger, "get_refined_top_points failed: ")
                return tobe_refined

        if USE_REFINED_BEFORE_CHK:
            # necessary for 20180329_microfluidic
            refined_cadidates = get_refined_top_points(top_cadidates)
            top_x = refined_cadidates

        else:
            top_x = top_cadidates

        if len(top_x) == 0:
            try:
                # as in 20181203
                n = np.argmax(spline(self.d1_roots))
                top_x = np.array([self.d1_roots[n]])
                self.get_logger()
                self.logger.warning("top_x has been made as [self.d1_roots[%d]] in an emergency measure.", n)
            except:
                pass

        if debug or len(top_x) == 0:
            import inspect
            for frm in inspect.stack()[1:]:
                print("---- : %s %s (%d)" % (frm.filename, frm.function, frm.lineno))

            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Tools.SpotDebugger import SpotDebugger
            debugger = SpotDebugger(None, (self,))
            debugger.show()

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
                fig.suptitle("No peak debug in " + __name__)
                ax1.plot(x, y)
                axt = ax2.twinx()
                ax2.plot(x, y, label='data')
                ax2.plot(x, self.spline_y, label='spline_y')
                ax2.plot(x, self.feature_y, label='feature_y')
                axt.plot(x, d1y, ':', label='d1y')
                axt.plot(x, d2y, ':', label='d2y')
                ax2.plot(self.d1_roots, spline(self.d1_roots), 'o', color='yellow', markersize=10, label='d1_roots')
                ax2.plot(sd1_roots, spline(sd1_roots), 'o', color='pink', label='sd1_roots')
                ax2.plot(top_cadidates, spline(top_cadidates), 'o', label='top_cadidates')
                ax2.plot(refined_cadidates, spline(refined_cadidates), 'o', label='refined_cadidates')
                # ax2.legend(loc='upper left')
                ax2.legend(loc='center right')
                axt.legend(loc='upper right')
                fig.tight_layout()
                plt.show()

        if len(top_x) == 0:
            raise RuntimeError("No peak")

        refined_x = get_refined_top_points(top_x)

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("compute_optimal_curves debug")
                ax.plot(x, self.spline(x))
                ax.plot(self.d1_roots, self.spline(self.d1_roots), 'o', color='yellow', markersize=15, alpha=0.5, label='d1_roots')
                ax.plot(top_cadidates, self.spline(top_cadidates), 'o', color='cyan', label='top_cadidates')
                if len(top_x) > 0:
                    ax.plot(top_x, self.spline(top_x), 'o', color='pink', label='top_x')
                    ax.plot(refined_x, self.spline(refined_x), 'o', color='red', label='refined_x')
                ax.legend()
                plt.show()

        self.top_x = np.array(refined_x)

    def baseline_correct_simply(self, x, y):
        k30 = int(len(y)*0.3)
        ppp = np.argpartition(y, k30)
        p30 = ppp[0:k30]
        x_ = x[p30]
        y_ = y[p30]
        slope, intercept, r_value, p_value, std_err = linregress( x_, y_ )
        corrected_y = y - (slope*x + intercept)
        if False:
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(x, y)
            ax.plot(x, corrected_y)
            ax.plot(x, slope*x + intercept, ':', color='red')
            plt.show()
        return corrected_y

    def get_original_top_x(self):
        ret_candidates = []
        if self.orig_top_x is None:
            pass
        else:
            refine_delta_x = REFINE_DELTA_X_RATIO * len(self.x)
            c_roots = self.d1_roots
            ret_candidates = []
            for cr in c_roots:
                dist = np.min(np.abs(self.orig_top_x - cr))
                if dist < refine_delta_x:
                    ret_candidates.append(cr)

            self.get_logger()
            self.logger.warning("top candidates %s have been retrieved from the original." % (str(ret_candidates)))
            if False:
                print('self.orig_top_x=', self.orig_top_x, 'refine_delta_x=', refine_delta_x)
                print('c_roots=', c_roots)
                print('ret_candidates=', ret_candidates)

        return np.array(ret_candidates)

    def get_good_top_candidates(self, debug=False):
        tall_enough = self.spline(self.d1_roots) - self.bottom > self.height*MAJOR_PEAK_RATIO
        max_d2 = np.max(np.abs(self.d2(self.d1_roots)))
        d2_is_well_negative = self.d2(self.d1_roots) < -max_d2 * NEGATIVE_D2_LIMIT_RATIO
        if False:
            print('d1_roots=', self.d1_roots)
            print('max_d2=', max_d2, max_d2 * NEGATIVE_D2_LIMIT_RATIO)
            print('d2(self.d1_roots=', d2(self.d1_roots))

        if debug:
            print('-------------- get_good_top_candidates begin')

        d1_roots = self.d1_roots
        num_roots = len(d1_roots)
        if num_roots < 2:
            return d1_roots

        default_delta_x = self.delta_x

        index = np.arange(num_roots)
        cadidates = np.logical_and( tall_enough, d2_is_well_negative )

        ret_candidates = []

        for k, root_x in enumerate(d1_roots):
            if not cadidates[k]:
                continue

            prev_r = d1_roots[k-1 if k > 0 else k+1]
            next_r = d1_roots[k+1 if k < num_roots - 1 else k-1]
            delta_r = min( abs(root_x - prev_r),  abs(next_r - root_x) )
            delta_r_ratio = delta_r/min(MAX_DENOM_DELTA_R_RATIO, len(self.x))

            prev_x = root_x - delta_r
            next_x = root_x + delta_r

            prev_y = self.spline_peak_recog(prev_x)
            next_y = self.spline_peak_recog(next_x)
            root_y = self.spline_peak_recog(root_x)

            delta_y = min(root_y - prev_y, root_y - next_y)
            delta_y_ratio = delta_y/self.height

            adopt = False
            if delta_y_ratio > self.min_y_ratio and delta_r_ratio > DELTA_X_LIMIT_RATIO:
                ret_candidates.append(root_x)
                adopt = True

            if debug:
                print([k], root_x, (delta_y_ratio, delta_r_ratio), adopt)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            print('-------------- get_good_top_candidates end')
            x = self.x
            y = self.y
            spline_pr = self.spline_peak_recog

            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(x, y)
            ax.plot(x, self.spline_peak_recog(x))
            ax.plot(self.d1_roots, spline_pr(self.d1_roots), 'o', label='d1_roots')
            ax.plot(ret_candidates, spline_pr(ret_candidates), 'o', label='ret_candidates')
            ax.legend()
            fig.tight_layout()
            plt.show()
            plt.pop()

        return np.array(ret_candidates)

    def get_halfmax_points_of_a_peak(self, k):
        half_h = (self.top_x[k] - self.bottom)/2
        # ...



# coding: utf-8
"""
    SimpleGuinierScoreDialog.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from scipy          import stats
from molass_legacy.KekLib.OurTkinter     import Tk, Dialog
from CanvasDialog   import CanvasDialog
from SimpleGuinier  import KNOT_NO_ADDITION
from SimpleGuinierScore import ( FWD_GUINIER_SIZE,
                                evaluate_rg_consistency,
                                evaluate_guinier_interval,
                                compute_end_consistency,
                                compute_fwd_consistency )

class SimpleGuinierScoreDialog( CanvasDialog ):
    def __init__( self, parent, title, guinier ):
        self.guinier    = guinier
        self.px         = guinier.px
        self.dx         = guinier.x[1] - guinier.x[0]
        self.x          = guinier.x2
        self.y          = guinier.log_y
        CanvasDialog.__init__(self, title, parent)

    def show( self, plot_title=None ):
        self.plot_title = plot_title
        CanvasDialog.show( self, self.draw, figsize=(12, 5), toolbar=True )

    def draw( self, fig ):
        self.fig = fig
        fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )
        self.point_list = []

        ax1 = fig.add_subplot( 121 )
        ax2 = fig.add_subplot( 122 )
        self.axes = [ ax1, ax2 ]
        self.redraw()

    def redraw( self, clear=False ):
        for i, ax in enumerate( self.axes ):
            if clear:
                ax.cla()
            if self.plot_title is not None:
                ax.set_title( self.plot_title )

            ax.set_xlabel( 'Q²' )
            ax.set_ylabel( 'Ln(I)' )

            ax.plot( self.guinier.x2, self.guinier.log_y, 'o', markersize=3 )
            ax.plot( self.guinier.x2, self.guinier.log_sy, color='green', alpha=0.5 )
            msize = 5 if i == 0 else 8
            ax.plot( self.guinier.kx2, self.guinier.log_ky, 'o', color='red', markersize=msize )
            if self.guinier.added_knot_type != KNOT_NO_ADDITION:
                ax.plot( self.guinier.kx2[1], self.guinier.log_ky[1], 'o', color='pink', markersize=msize )

            if i == 0:
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
            else:
                ax.set_xlim( xmin/16, xmax/64 )
                ymax_ = max( self.guinier.guinier_y[0]+0.5, self.guinier.log_py + 0.1 )
                ymax_ = max( ymax_, self.guinier.log_sy[0] + 0.1 )
                ax.set_ylim( self.guinier.guinier_y[1]-1.0, ymax_ )

            ax.plot( self.guinier.px2, self.guinier.log_py, 'o', color='yellow', markersize=msize )

        self.fig.tight_layout()

    def on_button_press( self, event ):

        if event.inaxes != self.axes[1]:
            return

        x, y = event.xdata, event.ydata
        # print( x, y )

        if len(self.point_list) == 2:
            del self.point_list[:]
            self.redraw( clear=True )

        ax = self.axes[1]
        dist = (self.x - x)**2 + (self.y - y)**2
        n = np.argmin( dist )
        point, = ax.plot( self.x[n], self.y[n], 'o', color='red' )
        self.point_list.append( (n, point) )

        if len( self.point_list ) == 2:
            n1 = self.point_list[0][0]
            n2 = self.point_list[1][0]
            i  = np.array( sorted( [n1,n2] ) )
            ax.plot( self.x[i], self.y[i], color='red' )

            tx = np.average( self.x[i] )
            ty = np.average( self.y[i] )
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            xoffset = (xmax-xmin)*0.05
            yoffset = (ymax-ymin)*0.05
            if ty > ymin*0.2 + ymax*0.8:
                yoffset *= -1
            elif ty > ymin*0.4 + ymax*0.6:
                yoffset = 0

            text = self.compute_score_text( i )
            ax.annotate( text, xy=( tx, ty ),
                xytext=( tx + 2*xoffset, ty + yoffset ), alpha=0.5,
                arrowprops=dict( headwidth=5, width=1, color='black', shrink=0, ),
                va='center',
                )

        self.mpl_canvas.show()

    def compute_score_text( self, i ):
        start = i[0]
        stop  = i[1]+1
        slice_ = slice( start, stop )
        x_ = self.x[slice_]
        y_ = self.y[slice_]
        slope, intercept, r_value, p_value, stderr = stats.linregress( x_, y_ )
        rg = np.sqrt( -3*slope )
        x  = np.sqrt( x_[-1] )
        size = stop - start
        end_consistency = end_consistency = compute_end_consistency( start, stop, self.x, self.y )
        fwd_consistency = compute_fwd_consistency( start, stop, self.x, self.y )
        score, score_vector, score_array = evaluate_guinier_interval( self.guinier.basic_quality, self.px, self.dx, rg , size, r_value, end_consistency, fwd_consistency, return_vector=True )

        return 'rg=%g\nq*rg=%g\nsize=%d\nsize_score=%g\nr_value=%g\nlinearity_score=%g\nfwd_consistency=%g\nend_consistency=%g\nscore=%g' % (
                    rg, x*rg, size, score_vector[0], r_value, score, fwd_consistency, end_consistency, score )



# coding: utf-8
"""

    DevSettings.py

    Copyright (c) 2017-2022, SAXS Team, KEK-PF
"""
from PersistentInfo     import PersistentInfo

settings_ = None

def dev_init(alt_folder=None):
    # 永続的な記憶をオブジェクト化する（メモリに取り込む）。
    global setting_info, settings_
    setting_file    = 'dev_settings.dump'
    setting_info    = PersistentInfo( setting_file, alt_folder=alt_folder )
    settings_       = setting_info.get_dictionary()
    # print("dev_init done", settings_)

ITEM_DEFAULTS = {
    'hiresolution'          : False,
    'take_screenshots'      : 0,
    'screenshot_folder'     : None,
    'intensity_reduction'   : 0,
    'reduction_method'      : 'THIN-OUT',
    'reduction_cycle'       : 4,
    'reduction_start'       : 0,
    'no_usable_q_limit'     : 0,
    'log_memory_usage'      : 0,
    'log_xray_lpm_params'   : 0,
    'show_num_iterations'   : 0,
    'suppress_defer_test'   : 0,
    'use_datgnom'           : 1,
    'use_simpleguinier'     : 1,
    'recompute_regboundary' : 0,
    'zx_add_constant'       : 0,
    'add_conc_const'        : 0,
    'individual_bq_ingore'  : 0,
    'tester_zx_save'        : 0,
    'adopt_lom'             : 0,
    'smoothed_xray_conc'    : 0,
    'enable_xb_save'        : 0,
    'make_excel_visible'    : 0,
    'keep_remaining_excel'  : 0,
    'enable_dnd_debug'      : 0,
    'xb_folder'             : None,
    'base_file_postfix'     : None,
    'running_with_tester'   : False,
    'preview_rg_list'       : None,
    }

def get_dev_setting( item ):
    global settings_
    if settings_ is None:
        dev_init()
    # print("get_dev_setting", settings_)
    assert( item in ITEM_DEFAULTS )
    value = settings_.get( item )
    if value is None:
        value = ITEM_DEFAULTS.get( item )
        set_dev_setting( item, value )
    return value

def set_dev_setting( item, value ):
    global settings_
    assert( item in ITEM_DEFAULTS )
    settings_[ item ] = value



# coding: utf-8
"""

    MultiMonitor.py

    Copyright (c) 2019-2021, Masatsuyo Takahashi, KEK-PF

"""
import os
from ctypes import windll, Structure, c_long, byref

PYTHON_DEMO_MONITOR     = 'PYTHON_DEMO_MONITOR'

"""
learned at
https://stackoverflow.com/questions/33685534/trying-to-retrieve-the-active-monitor-id-in-python/33760154
"""

"""
from
    Getting cursor position in Python
    https://stackoverflow.com/questions/3698635/getting-cursor-position-in-python

    changed from c_ulong to c_long
    to handle cases where the second monitor
    is placed to the left side of the first.
"""
class POINT(Structure):
    _fields_ = [("x", c_long), ("y", c_long)]

def get_mouse_position():
    pt = POINT()
    windll.user32.GetCursorPos(byref(pt))
    return pt.x, pt.y

monitors = None
max_monitor = None

def get_monitor_list():
    global monitors
    if monitors is None:
        try:
            from screeninfo import get_monitors
            monitors = get_monitors()
        except:
            print( 'screeninfo is not installed.' )
            monitors = []
    return monitors

def get_max_monitor():
    global monitors, max_monitor

    if max_monitor is not None: return max_monitor

    monitors = get_monitor_list()

    for m in monitors:
        # print( 'get_max_monitor', str( m ) )
        if max_monitor is None or m.width + m.height > max_monitor.width + max_monitor.height:
            max_monitor = m

    return max_monitor


selected_monitor = None

def get_selected_monitor():
    global selected_monitor

    if selected_monitor is not None:
        return selected_monitor

    selected_monitor = get_max_monitor()
    demo_monitor = os.environ.get( PYTHON_DEMO_MONITOR )
    # print( 'demo_monitor=', demo_monitor )
    if demo_monitor is None:
        from MultiMonitor import get_mouse_position
        x, y = get_mouse_position()
        for m in monitors:
            # print(m.x, m.y, m.width)
            if x >= m.x and x <= m.x + m.width:
                selected_monitor = m
                break
    else:
        selected_monitor = monitors[ int(demo_monitor) ]
        # print( 'selected monitor=', monitor )
    return selected_monitor



"""
    SuperEditorFrame.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import os
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.gridspec    import GridSpec
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib import colors
import matplotlib
import seaborn as sns
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import split_geometry
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar, get_color, get_hex_color
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting
from DevSettings            import get_dev_setting
from molass_legacy.DataStructure.AnalysisRangeInfo      import AnalysisRangeInfo, convert_to_paired_ranges
from .DecompSpecPanel import RANGE_FRAME_HEIGHT
from molass_legacy._MOLASS.Version import get_version_string

MIN_NUM_PANEL_SPACES    = 3
CANVAS_HIGHT    = 6

def format_coord(x, y):
    return 'x=%.4g    y=%.4g' % (x, y)

class SuperEditorFrame(Tk.Frame):
    def __init__( self, parent, title):
        Tk.Frame.__init__(self, parent)

        self.build_body(self, title)
        # self.create_popup_menu()  # deprecated?

    def close_figs(self):
        for fig in [self.fig1, self.fig2]:
            plt.close(fig)

    def build_body(self, body_frame, title):

        title_frame = Tk.Frame( body_frame )
        title_frame.pack()
        label = Tk.Label(title_frame, text=title, font=("", 20) )
        label.pack(pady=5)

        self.panel_frame = panel_frame = Tk.Frame( body_frame )
        panel_frame.pack()

        button_frame = Tk.Frame( body_frame )
        button_frame.pack(fill=Tk.X, expand=1)

        ctframe = Tk.Frame( panel_frame )
        ctframe.grid(row=0, column=0, sticky=Tk.W+Tk.E)
        self.data_label = Tk.Label(ctframe)
        self.set_data_label()
        self.data_label.pack(side=Tk.LEFT)
        toolbar_frame = Tk.Frame(ctframe)
        toolbar_frame.pack(side=Tk.LEFT, fill=Tk.X, expand=1, padx=10)

        ccframe = Tk.Frame( panel_frame )
        ccframe.grid(row=1, column=0)

        self.ptframe = Tk.Frame( panel_frame )
        self.ptframe.grid(row=0, column=1, sticky=Tk.W+Tk.E)
        p_frame = Tk.Frame( panel_frame )
        p_frame.grid(row=1, column=1, sticky=Tk.W+Tk.E+Tk.N+Tk.S)

        self.fig1 = fig1 = plt.figure( figsize=(7, CANVAS_HIGHT) )
        self.mpl_canvas1 = FigureCanvasTkAgg( fig1, ccframe )
        self.mpl_canvas1_widget = self.mpl_canvas1.get_tk_widget()
        self.mpl_canvas1_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas1.mpl_connect('button_press_event', self.on_mpl_button_press)

        self.canvas_draw1()
        height_ratios = [5, 1]
        gs = GridSpec( 2, 1, height_ratios=height_ratios )
        self.height_ratio = height_ratios[1]/height_ratios[0]

        self.ax1 = fig1.add_subplot(gs[0,0])
        self.ax1.get_xaxis().set_visible(False)
        self.ax1r = fig1.add_subplot(gs[1,0], sharex=self.ax1)

        for ax in [self.ax1, self.ax1r]:
            ax.format_coord = format_coord    # override the default to avoid fluctuating

        self.draw1()
        fig1.tight_layout()
        # self.toolbar = NavigationToolbar( self.mpl_canvas1, toolbar_frame, show_mode=False )
        self.toolbar = NavigationToolbar( self.mpl_canvas1, toolbar_frame )
        self.toolbar.update()

        self.canvas1_height = int( self.mpl_canvas1_widget.cget( 'height' ) )
        pheight_adjust = Tk.Frame( p_frame, height=self.canvas1_height )
        pheight_adjust.grid(row=0, column=0 )

        self.pcframe_padx = 10
        self.pcframe = Tk.Frame( p_frame )
        self.pcframe.grid(row=0, column=0, padx=self.pcframe_padx, sticky=Tk.N)     # simply adding +Tk.S can be a problem when the figure canvas is resized
                                                                                    # see also adjust_the_tkframe_size  
        self.ppframe = Tk.Frame( p_frame )
        self.ppframe.grid(row=0, column=1, sticky=Tk.N+Tk.S)

        figsize = self.get_figsize()
        self.fig2 = fig2 = plt.figure( figsize=figsize )
        self.mpl_canvas2 = FigureCanvasTkAgg( fig2, self.pcframe )
        self.mpl_canvas2_widget = self.mpl_canvas2.get_tk_widget()
        self.mpl_canvas2_widget.pack( fill=Tk.BOTH, expand=1 )
        self.canvas_draw2()

        self.axes = None
        self.draw2()
        fig2.tight_layout()
        self.update()
        self.add_widgets()

        for j in range(2):
            button_frame.grid_columnconfigure(j, weight=1)

        button_frame_left = Tk.Frame(button_frame)
        button_frame_left.grid( row=0, column=0 )

        if self.toggle:
            self.toggle_btn = Tk.Button(button_frame_left, command=self.toggle_show )
            self.set_toggle_text()
            self.toggle_btn.pack( side=Tk.LEFT)

        self.error_label = Tk.Label(button_frame_left)
        self.error_label.pack( side=Tk.LEFT, padx=40 )
        self.update_error_label()

        button_frame_right = Tk.Frame(button_frame)
        button_frame_right.grid( row=0, column=1, sticky=Tk.E )

        # if get_version_string().find("dev") > 0:
        if False:
            if get_setting('enable_debug_plot'):
                debug_btn = Tk.Button(button_frame_right, text="Debug Plot", command=self.debug_plot )
                debug_btn.grid( row=0, column=0, padx=50 )

            reset_btn = Tk.Button(button_frame_right, text="Reset to Defaults", state=Tk.DISABLED, command=self.reset_to_defaults )
            reset_btn.grid( row=0, column=1, padx=50 )

            if  self.params_controllable:
                self.param_btn = Tk.Button(button_frame_right, text="Show Parameter Constraints", command=self.toggle_params_constraints )
                self.param_btn.grid( row=0, column=2, padx=5 )
                self.recompute_btn = Tk.Button(button_frame_right, text="Recompute", command=self.recompute_decomposition )
                self.recompute_btn.grid( row=0, column=3, padx=5 )
                self.recompute_btn.grid_forget()

    def debug_plot(self):
        # overrided in DecompEditorFrame or RangeEditorFrame
        from importlib import reload
        from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo
        import DecompEditorDebug
        reload(DecompEditorDebug)
        from DecompEditorDebug import debug_plot_impl
        ranges = AnalysisRangeInfo(self.make_range_info()).get_ranges()
        debug_plot_impl(self, self.dialog)

    def get_figsize(self):
        fig_height = min( CANVAS_HIGHT, self.num_eltns*2 )
        return (3, fig_height)

    def update_error_label(self):
        text='fit_error=%.3g' % self.fit_error
        self.error_label.config(text=text)

    def canvas_draw1(self):
        self.mpl_canvas1.draw()

    def canvas_draw2(self):
        self.mpl_canvas2.draw()

    def draw1(self):
        pass

    def draw2(self, peak_no=None):
        pass

    def add_range_patchs(self, ax, range_parts, k=None):
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        range_parts.append([ymin, ymax])

        if k is None:
            editor_ranges_ = self.editor_ranges
        else:
            editor_ranges_ = [self.editor_ranges[k]]

        for j, range_list in enumerate(editor_ranges_):
            if k is None and self.ignorable_flags[j]:
                range_parts.append([j])
                continue

            for f, t in range_list:
                p = mpl_patches.Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax.add_patch( p )
                lines = []
                for x in [f, t]:
                    line, = ax.plot( [x, x], [ymin, ymax], ':', color='gray', alpha=0.5 )
                    lines.append(line)
                range_parts.append([j, p, lines])

    def refresh_figs(self):
        self.draw1()
        self.draw2()
        self.canvas_draw1()
        self.canvas_draw2()

    def update1(self):
        print('update1 start')
        try:
            self.update1_impl(self.fig1_range_parts)
            self.update1_impl(self.fig1_range_parts_resid)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
        print('update1 end')

    def update1_impl(self, range_parts):
        print('update1_impl start')
        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            print([k, j], 'range_list=', range_list)
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)
                for n, x_ in enumerate([f, t]):
                    line = lines[n]
                    line.set_data([x_, x_], [ymin, ymax])

    def update2(self, peak_no):
        print('update2: peak_no=', peak_no)
        pk = peakno - 1
        range_parts = self.fig2_range_parts_list[pk]

        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)

    def update_figs(self, peak_no):
        self.update1()
        self.update2(peak_no)
        self.canvas_draw1()
        self.canvas_draw2()

    def add_widgets(self):
        num_eltns = self.num_eltns

        canvas_width = int(self.mpl_canvas2_widget.cget('width')) + self.pcframe_padx*2
        self.col_widths  = [canvas_width, 60, 120, 80, 60, 140, 60, 60, 160]
        # self.col_widths[2] = 120 required for SUB_TRN1 with EMG
        label_texts = [ 'Figure',
                        'Peak No',
                        'Element Curve(s)',
                        'Number of\nRanges',
                        'Range Id',
                        'Range(s)\nFrom       To',
                        'Ignore',
                        ]
        if self.params_controllable:
            label_texts.append( 'τ Value' )
            label_texts.append( 'τ Constraints' )

        self.col_title_frames = []
        for k, text in enumerate(label_texts):
            frame = Tk.Frame(self.ptframe, width=self.col_widths[k], height=40)
            frame.grid(row=0, column=k, sticky=Tk.S)
            self.col_title_frames.append(frame)
            # https://stackoverflow.com/questions/16363292/label-width-in-tkinter
            frame.pack_propagate(0)
            label = Tk.Label(frame, text=text)
            label.pack(side=Tk.BOTTOM)

        if self.params_controllable:
            self.col_title_frames[-2].grid_forget()
            self.col_title_frames[-1].grid_forget()

        for k in range(max(MIN_NUM_PANEL_SPACES, num_eltns)):
            self.ppframe.grid_rowconfigure(k, weight=1)

        if self.is_memorized and RESTORE_HINTS_DICT:
            hints_dict = get_setting('tau_hints_dict')
            self.logger.info('restored hints_dict will be used: ' + str(hints_dict))
        else:
            hints_dict = None
        self.refresh_spec_panels(hints_dict=hints_dict)

        for k in range( max(0, MIN_NUM_PANEL_SPACES - num_eltns) ):
            space = Tk.Frame(self.ppframe, height=RANGE_FRAME_HEIGHT)
            space.grid(row=num_eltns+k, column=0)

    def set_data_label(self):
        data_type = "Xray" if self.elution_fig_type == 0 else "UV"
        self.data_label.config( text=data_type + " elution decomposition" )

    def set_toggle_text(self):
        # override
        pass

    def toggle_show( self ):
        self.elution_fig_type = 1 - self.elution_fig_type
        self.refresh_frame()

    def refresh_frame(self, elution_fig_type=None):
        if elution_fig_type is not None:
            self.elution_fig_type = elution_fig_type
        self.set_data_label()
        self.set_toggle_text()
        self.refresh_figs()
        self.update_error_label()

    def reset_to_defaults( self ):
        pass

    def get_fit_error( self ):
        return self.fit_error



# coding: utf-8
"""
    BaseEstimater.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt 

"""
    
"""
class BaseEstimater:
    def __init__(self, D, E=None, base_type=None):
        self.base_plane = np.zeros(D.shape)

    def get_baseline(self, i):
        return self.base_plane[i,:]

def demo(in_folder):
    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from MatrixData import simple_plot_3d

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()

    U, _, wv, uv_curve = sd.get_uv_data_separate_ly()
    # print(U.shape, wv.shape, uv_curve.x.shape)

    q1 = get_setting('intensity_picking')
    q2 = q1*2
    w1 = get_setting('absorbance_picking_sub')
    w2 = get_setting('absorbance_picking')

    for curve, D_, vector, v1, v2, vname, colors in [
                                (xr_curve, D, qv, q1, q2, 'q', ('orange', 'green')),
                                (uv_curve, U, wv, w1, w2, 'wavelength', ('cyan', 'blue')),
                                ]:

        x = curve.x
        y = curve.y

        i1 = bisect_right(vector, v1)
        i2 = bisect_right(vector, v2)

        be = BaseEstimater(D_)
        b1 = be.get_baseline(i1)
        b2 = be.get_baseline(i2)

        plt.push()
        fig = plt.figure(figsize=(21, 7))
        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        fig.suptitle("Baseline Estimation using Decomposition-Rg Consistency", fontsize=24)
        ax1.set_title("3D View", fontsize=20)
        ax2.set_title("Elution at %s=%.03g" % (vname, v1), fontsize=20)
        ax3.set_title("Elution at %s=%.03g" % (vname, v2), fontsize=20)

        simple_plot_3d(ax1, D_, x=vector)
        y_ = x

        y2 = np.average(D_[i2-5:i2+6,:], axis=0)

        for v_, z_, color in [(v1, y, colors[0]), (v2, y2, colors[1])]:
            x_ = np.ones(len(x))*v_
            ax1.plot(x_, y_, z_, color=color)

        ax2.plot(x, y, color=colors[0])
        ax2.plot(x, b1, ':', color='red')

        ax3.plot(x, y2, color=colors[1])
        ax3.plot(x, b2, ':', color='red')
        ymin2, ymax2 = ax2.get_ylim()
        ymin3, ymax3 = ax3.get_ylim()
        ymin = min(ymin2, ymin3)
        ymax = max(ymax2, ymax3)
        for ax in [ax2, ax3]:
            ax.set_ylim(ymin, ymax)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
        plt.show()
        plt.pop()



# coding: utf-8
"""
    NoExcelWarningDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import re
import os
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

class NoExcelWarningDialog( Dialog ):
    def __init__(self, parent, com_error):
        self.parent = parent
        self.com_error = com_error
        Dialog.__init__( self, parent, "No Excel Warning", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        self.ro_text = ReadOnlyText(body_frame, height=6, relief=Tk.FLAT, fg='orange')
        self.ro_text.pack(padx=20, pady=10)

        warning_massage = (
            "No Excel instance is available in this environment for some reason\n"
            "such as no Excel has yet been properly installed.\n\n"
            "Be aware that some formatting will be skipped in the analysis result book\n"
            "to leave it look a little degraded.\n"
            )
        self.ro_text.insert(Tk.END, warning_massage)

        if self.com_error is not None:
            ro_text = ReadOnlyText(body_frame, height=20, relief=Tk.FLAT, fg='red')
            ro_text.pack(padx=20, pady=10)
            ro_text.insert(Tk.END, self.com_error )

    def buttonbox(self):

        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Do not repeat this message", width=24, command=self.do_not_repeat)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def do_not_repeat(self):
        set_setting('no_excel_warning', False)
        self.cancel()



"""
    Optimizer.FvSynthesizerHistory.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

score_names = [
    "XR_2D_fitting",
    "XR_LRF_residual",
    "UV_2D_fitting",
    "UV_LRF_residual",
    "Guinier_deviation",
    "Kratky_smoothness",
    "SEC_conformance",
    ]

weights_history = [
    [0.11111111, 0.22222222, 0.11111111, 0.22222222, 0.11111111, 0.11111111, 0.11111111],
    [0.2, 0.1, 0.2, 0.1, 0.133, 0.133, 0.133],
    [0.3, 0.1, 0.3, 0.1, 0.067, 0.067, 0.066],
    [0.35, 0.1, 0.35, 0.1, 0.033, 0.033, 0.034],
    ]

def plot_history():
    ncols = len(weights_history)
    xpos = np.arange(len(score_names))
    colors = ["C1", "C0", "C1", "C0", "C0", "C0", "C0"]
    fig, axes = plt.subplots(ncols=ncols, figsize=(12,4))
    fig.suptitle("Tried Weights Variations")
    trial = 0
    for ax, weights in zip(axes, weights_history):
        trial += 1
        ax.set_title("Trial %d" % trial)
        ax.bar(score_names, weights, color=colors)
        ax.set_xticks(xpos, score_names, rotation=90)
        ax.set_ylim(0, 0.5)

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    import seaborn
    seaborn.set()
    plot_history()



# coding: utf-8
"""
    SliceUtils.py

    Copyright (c) 2021, Masatsuyo Takahashi, KEK-PF
"""

import numpy as np

"""
    from 
    https://stackoverflow.com/questions/53361203/numpy-efficient-way-to-extract-range-of-consecutive-numbers
"""
def slice_consecutives(ary):
  pass
  res=[]
  tmp = [ary[0]]
  for idx, x in np.ndenumerate(ary[1:]):
    if x - ary[idx[0]] > 1 or idx[0] + 2 == len(ary):
      if tmp[0] != tmp[-1]: res.append([tmp[0], tmp[-1]])
      tmp = []
    tmp.append(x)
  if ary[-1] - res[-1][-1] == 1: res[-1][-1] = ary[-1]
  return res



"""
    TkDebugDnd.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog
import tkinterDnD

class TkDebugDnd(Dialog):
    def __init__(self, parent, nrows=1, ncols=1):
        self.nrows = nrows
        self.ncols = ncols
        Dialog.__init__(self, parent, title="TkDebugDnd")

    def body(self, frame):
        for i in range(self.nrows):
            for j in range(self.ncols):
                label = Tk.Label(frame, text=str((i,j)), width=20, height=2, relief=Tk.RIDGE)
                label.grid(row=i, column=j)
                label.register_drop_target("*")
                label.bind("<<Drop>>", lambda event, i_=i, j_=j: self.on_drop(event, i_, j_))

                label.register_drag_source("*")
                label.bind("<<DragInitCmd>>", lambda event, i_=i, j_=j: self.on_drag(event, i_, j_))

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def on_drop(self, event, i, j):
        print("on_drop", (i,j), event.data)

    def on_drag(self, event, i, j):
        print("on_drag")
        return (tkinterDnD.COPY, "DND_Text", "on_drag text: " + str((i,j)))




"""
    SecTheory.MonoPore.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from .BasicModels import robust_single_pore_pdf_scaled
import molass_legacy.SecTheory.CumsumInverse
from importlib import reload
reload(molass_legacy.SecTheory.CumsumInverse)
import molass_legacy.SecTheory.LocalOptimizer
reload(molass_legacy.SecTheory.LocalOptimizer)
from .LocalOptimizer import PENALTY_SCALE, MEMP_LIMIT, N_LIMIT, RgInfo, SecInfo, LocalOptimizer, split_params
# MEMP_LIMIT and N_LIMIT are imported using this module although they are not used here.

def estimate_monopore_params(ecurve, rg_curve, nc, optimizer=None,
        lrf_src=None,
        elutionmodel_func=robust_single_pore_pdf_scaled,
        t0_upper_bound=None,
        global_opt=False, init_params=None, init_seccol_params=None,
        logger=None,
        debug=False,
        just_get_curves=False):
    """
    to estimate t0, rp, N, me, T, mp

        rho = rho = 1 if rg > rp else rg/rp
        np_ = N*(1 - rho)**me
        tp_ = T*(1 - rho)**mp
        Ksec = (1 - rho)**(me + mp)
        tR = t0 + T*Ksec

    Rg → tR

    """
    if debug:
        import molass_legacy.SecTheory.InitialGuess
        reload(molass_legacy.SecTheory.InitialGuess)
    from .InitialGuess import InitialGuess

    rg_info = RgInfo(rg_curve, nc)
    sec_info = SecInfo(t0_upper_bound=t0_upper_bound)
    localopt = LocalOptimizer(ecurve, rg_info, sec_info, nc, elutionmodel_func, logger=logger)

    params_given = True
    if init_params is None:
        params_given = False
        try:
            initguess = InitialGuess(ecurve, rg_info, sec_info, localopt, debug=debug)
            init_params = initguess.get_params()
            init_tr = initguess.init_tr
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(logger, "InitialGuess failed: ")
            return
    else:
        (t0, rp, N, me, T, mp), rg, w = split_params(init_params, nc)
        rho = rg/rp
        rho[rho > 1] = 1
        init_tr = t0 + N*T*(1 - rho)**(me + mp)

    if just_get_curves:
        return localopt.objective_all(init_params, get_curves=True)

    ret = localopt.optimize(init_params, params_given=params_given, init_tr=init_tr, global_opt=global_opt, debug=debug)

    return ret

def compute_rg(t0, rp, N, me, T, mp, tr):
    from scipy.optimize import root

    def fun(x):
        return t0 + N*T*(1 - x)**(me + mp) - tr

    ret = root(fun, np.ones(len(tr))*0.7)
    return ret.x * rp

def compute_mu_sigma(t0, rp, N, me, T, mp, rg):
    rho = 1 if rg > rp else rg/rp
    np_ = N*(1 - rho)**me
    tp_ = T*(1 - rho)**mp
    return t0 + np_*tp_, np.sqrt(2 * np_ * tp_**2)

def get_modified_params(ecurve, nc, params, elutionmodel_func=robust_single_pore_pdf_scaled, logger=None, debug=False):
    from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks, egh

    x = ecurve.x
    y = ecurve.y

    (t0, rp, N, me, T, mp), rg, w = split_params(params, nc)
    rho = rg/rp
    rho[rho > 1] = 1
    ty = np.zeros(len(x))

    model_trs = t0 + N*T*(1 - rho)**(me + mp)

    penalty = 0
    preference = 0
    cy_list = []
    for w_, r_ in zip(w, rho):
        penalty += min(0, w_)**2
        np_ = N*(1 - r_)**me
        tp_ = T*(1 - r_)**mp
        cy = w_*elutionmodel_func(x - t0, np_, tp_)
        ty += cy
        cy_list.append(cy)

    resid_y = np.max([np.zeros(len(x)), y - ty], axis=0)
    for L, _, R in ecurve.peak_info:
        resid_y[L:R+1] = 0

    total_area = np.sum(ty)
    try:
        peaks = recognize_peaks(x, resid_y, num_peaks=2)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(logger, "recognize_peaks: ")
        return None

    w_pairs = [(i, v) for i, v in zip(np.arange(nc), w)]
    s_pairs = sorted(w_pairs, key=lambda x: x[1])

    i = 0
    modify_list = []
    for k, (h, m, s, t) in enumerate(peaks):
        ry = egh(x, h, m, s, t)
        area_ratio = np.sum(ry)/total_area
        print([k], "area_ratio=", area_ratio)
        if area_ratio > 0.01:
            modify_list.append((s_pairs[i][0], m))
            i += 1

    if len(modify_list) == 0:
        return None

    modify_array = np.array(modify_list)
    rgs_to_add = compute_rg(t0, rp, N, me, T, mp, modify_array[:,1])

    new_rgw_list = []
    i = 0
    for k, (rg_, w_) in enumerate(zip(rg, w)):
        if i < modify_array.shape[0]:
            j = int(modify_array[i,0])
        else:
            j = -1
        print([k], i, j)
        if k == j:
            print("append", i)
            new_rgw_list.append((rgs_to_add[i], w_))
            i += 1
        else:
            new_rgw_list.append((rg_, w_))

    new_rgw_list = sorted(new_rgw_list, key=lambda x: -x[0])    # decending order of Rg
    new_rgw_array = np.array(new_rgw_list)
    new_rg = new_rgw_array[:,0]
    new_w = new_rgw_array[:,1]

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax1.plot(x, y)
            for cy, tr in zip(cy_list, model_trs):
                ax1.plot(x, cy, ":")
                j = int(round(tr))
                if j < len(cy):
                    py = cy[j]
                    ax1.plot(tr, py, "o", color="yellow")
            ax1.plot(x, ty, ":", color="red")
            ax2.plot(x, resid_y)
            for peak in peaks:
                ax2.plot(x, egh(x, *peak), ":")
            fig.tight_layout()
            plt.show()

    return np.concatenate([(t0, rp, N, me, T, mp), new_rg, new_w])

def estimate_uv_scale_params(xr_curve, rg_curve, uv_curve, nc, init_params, xr_w, uv_x, uv_y, xr_x, xr_y, optimizer=None, debug=False):
    # refactoring memo: too many arguments. better make it to a class

    cy_list, ty = estimate_monopore_params(xr_curve, rg_curve, nc, init_params=init_params, optimizer=optimizer, just_get_curves=True, debug=debug)
    uv_w = xr_w/xr_curve.max_y * uv_curve.max_y

    pd_cy_list = [cy/w for cy, w in zip(cy_list, xr_w)]

    # optimize uv_w
    def objective_func(p):
        ty = np.zeros(len(uv_y))
        negetive_penalty = 0
        for w, pd_cy in zip(p, pd_cy_list):
            negetive_penalty += min(0, w)**2
            ty += w*pd_cy
        negetive_penalty *= PENALTY_SCALE
        return np.sum((ty - uv_y)**2) + negetive_penalty

    ret_uv = minimize(objective_func, uv_w)
    uv_w = ret_uv.x

    baseline_type = get_setting("unified_baseline_type")
    if baseline_type == 2:
        uv_ty = np.zeros(len(uv_x))
        for k, pd_cy in enumerate(pd_cy_list):
            uv_cy = pd_cy*uv_w[k]
            uv_ty += uv_cy
    else:
        uv_ty = None    # not used in linear baseline

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("estimate_uv_scale_params")
            ax1.plot(uv_x, uv_y, color="blue", label="data")
            ax2.plot(xr_x, xr_y, color="orange", label="data")
            uv_ty = np.zeros(len(uv_x))
            for k, pd_cy in enumerate(pd_cy_list):
                uv_cy = pd_cy*uv_w[k]
                uv_ty += uv_cy
                ax1.plot(uv_x, uv_cy, ":", label="component-%d" % (k+1))
                xr_cy = pd_cy*xr_w[k]
                ax2.plot(xr_x, xr_cy, ":", label="component-%d" % (k+1))

            ax1.plot(uv_x, uv_ty, ":", color="red", label="component-total")
            ax2.plot(xr_x, ty, ":", color="red", label="component-total")
            ax1.legend()
            ax2.legend()
            fig.tight_layout()
            plt.show()

    return uv_w, uv_ty



"""
    ABC.HistoryUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from .ParameterArray import ParameterArray

def get_min_dist_parameter(history, return_values=False):
    df, w = history.get_distribution(m=0, t=history.max_t)
    population = history.get_population(history.max_t)
    distance_df = population.get_weighted_distances()
    k = np.argmin(distance_df['distance'].values)
    values = df.iloc[k].values
    if return_values:
        return values
    return ParameterArray(values)


"""
    Qmm3dFigure.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from mpl_toolkits.mplot3d import Axes3D
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MatrixData import simple_plot_3d

COLOR_ZOOMED = 'yellow'

class Qmm3dFigure(Dialog):
    def __init__(self, parent, dataset, frame):
        self.dataset = dataset
        self.frame = frame
        Dialog.__init__(self, parent, title="QMM 3d Figure", visible=False)

    def show(self):
        self._show()

    def body(self, bframe):

        fig = plt.figure(figsize=(16,7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, bframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122, projection='3d')

        self.find_where_to_zoom()
        self.draw_whole(ax1)
        self.draw_zoomed(ax2)

        fig.tight_layout()
        self.mpl_canvas.draw()

    def draw_whole(self, ax):
        ax.set_title("Linear Plot of the Whole Region", fontsize=16)
        self.dataset.plot_3d(ax, alpha=0.3)
        simple_plot_3d(ax, self.zoom_data, x=self.zoom_x, y=self.zoom_y, color=COLOR_ZOOMED, edgecolors=COLOR_ZOOMED)
        ax.set_xlabel("$Q$")
        ax.set_ylabel("$Eno$")
        ax.set_zlabel("$Intensity$")

    def draw_zoomed(self, ax):
        ax.set_title("Guinier Plot of the Small Angle Peak Region", fontsize=16)
        data = np.log(self.zoom_data)
        y = self.zoom_y
        px = self.zoom_x**2
        simple_plot_3d(ax, data, x=px, y=y, color=COLOR_ZOOMED, alpha=0.2)

        frame = self.frame
        gy = frame.gy_list[0]
        mu = frame.mu_list[0]

        aslice = self.zoom_info[0]
        e11n = frame.result_list[0]  # e11n
        for k in self.zoom_w:
            m = int(round(mu[k]))
            py = np.ones(len(px))*m
            z = e11n.P[aslice, k]
            pz = np.log(z*gy[m])
            ax.plot(px, py, pz, label='c-%d' % k, color='C%d' % (k+2))

        ax.legend()
        ax.set_xlabel("$Q^2$")
        ax.set_ylabel("$Eno$")
        ax.set_zlabel("$Ln(Intensity)$")

    def find_where_to_zoom(self):
        self.zoom_info = find_where_to_zoom_impl(self.dataset, self.frame)
        aslice, eslice, w = self.zoom_info
        self.zoom_data = self.dataset.data[aslice, eslice]
        self.zoom_x = self.dataset.vector[aslice]
        self.zoom_y = np.arange(eslice.start, eslice.stop)
        self.zoom_w = w

def find_where_to_zoom_impl(dataset, frame):
    astart = 0
    x = dataset.vector
    astop = bisect_right(x, 0.04)

    gy = frame.gy_list[0]
    C = frame.C_list[0]
    mu = frame.mu_list[0]
    sigma = frame.sigma_list[0]
    max_y = np.max(gy)
    n = len(mu)
    flags = np.zeros(n, dtype=int)
    for k, c in enumerate(C[:n]):
        m = int(mu[k])
        ratio = c[m]/max_y
        flags[k] = round(ratio)

    w = np.where(flags==1)[0]
    w0 = w[0]
    estart = int(round(mu[w0] - sigma[w0]))
    w1 = w[-1]
    estop = int(round(mu[w1] + sigma[w1])) + 1
    print('flags=', flags, w, estart, estop)
    return slice(astart, astop), slice(estart, estop), w



# coding: utf-8
"""

    WindowSizeTester.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import sys
import os
import time
from pyautogui              import screenshot
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from molass_legacy.KekLib.BasicUtils             import mkdirs_with_retry, clear_dirs_with_retry
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog, Font
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.TkUtils                import adjusted_geometry, split_geometry, join_geometry, get_max_monitor
import OurMessageBox        as MessageBox
from ChangeableLogger       import Logger, arg_join

class TestDialog( Dialog ):
    def __init__( self, parent, title, width, height ):
        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = title
        self.width      = width
        self.height     = height
        self.fixed_font = Font.Font( family="Courier", size=32 )

    def show( self ):
        Dialog.__init__( self, self.parent, self.title_, auto_geometry=False ) # this calls body

    def body( self, body_frame ):
        body_frame.pack( fill=Tk.BOTH, expand=Tk.Y )

        frame = Tk.Frame( body_frame, width=self.width, height=self.height )
        frame.pack( fill=Tk.BOTH, expand=Tk.Y )

        if self.parent.auto:
            escape_label = Tk.Label( frame, text='Press <escape> to stop' )
            escape_label.grid( row=1, column=1 )

        label_nw = Tk.Label( frame, text='NW', font=self.fixed_font )
        label_nw.grid( row=0, column=0, sticky=Tk.NW )
        label_ne = Tk.Label( frame, text='NE', font=self.fixed_font )
        label_ne.grid( row=0, column=2, sticky=Tk.NE )
        label_sw = Tk.Label( frame, text='SW', font=self.fixed_font )
        label_sw.grid( row=2, column=0, sticky=Tk.SW )
        label_se = Tk.Label( frame, text='SE', font=self.fixed_font )
        label_se.grid( row=2, column=2, sticky=Tk.SE )

        for x in range(3):
          Tk.Grid.columnconfigure(frame, x, weight=1)
        for y in range(3):
          Tk.Grid.rowconfigure(frame, y, weight=1)

        # W, H, X, Y = split_geometry( self.parent.geometry() )
        self.x = self.parent.X + self.parent.W // 2 - self.width // 2
        self.y = self.parent.Y + self.parent.H // 2 - self.height // 2
        geom = join_geometry( self.width, self.height, self.x, self.y )
        print( 'geom=', geom )
        self.geometry( geom )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="Shot", width=10, command=self.shot, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def shot( self ):
        file = os.path.join( self.parent.img_folder, self.title_ + '.png' )
        # screenshot( file, region=( self.x, self.y, self.width, self.height) )
        screenshot( file )


class WindowSizeTester( Tk.Toplevel ):
    def __init__( self, parent ):
        Tk.Toplevel.__init__( self, parent )

        parent.report_callback_exception = self.report_callback_exception
        self.parent = parent
        self.auto   = False

        wrk_folder = __file__
        for i in range(3):
            wrk_folder = os.path.dirname( wrk_folder )
        log_folder= wrk_folder + r'\log'
        if not os.path.exists( log_folder ):
            mkdirs_with_retry( log_folder )

        log_file = log_folder + r'\test.log'
        if os.path.exists( log_file ):
            os.remove( log_file )

        self.app_logger = Logger( log_file )
        img_folder = wrk_folder + r'\img'
        clear_dirs_with_retry( [ img_folder ] )
        self.img_folder = img_folder

        self.withdraw()
        self.build_window()
        self.update()

        max_monitor = get_max_monitor()
        print( 'max_monitor=', max_monitor )
        W, H, X, Y = max_monitor.width, max_monitor.height, max_monitor.x, max_monitor.y
        self.geometry( join_geometry( 920, 650, X + 100, Y + 100 ) )
        self.W  = W     # 1920
        self.H  = H     # 1080
        self.X  = X
        self.Y  = Y

        self.W_min  = 240
        self.W_inc  = ( self.W - self.W_min ) / 9
        self.H_min  = 180
        self.H_inc  = ( self.H - self.H_min ) / 9

        # self.resizable(width=False, height=False)
        self.deiconify()
        self.bind( '<Escape>', self._on_escape )

    def report_callback_exception(self, exc, val, tb):
        # This method is to override the Tk method to be able
        # to report to the spawned console in windows application mode.
        et = ExceptionTracebacker()
        msg = 'Overridden Tk report method: ' + str( et )
        if self.app_logger is None:
            print( msg )
        else:
            self.app_logger.error( msg )

    def build_window( self ):
        global radio_button_text_length

        tk_set_icon_portable( self )

        label = Tk.Label( self, text="This is a window size tester." )
        label.pack( pady=30 )
        test_frame = Tk.Frame( self )
        # test_frame.pack( anchor=Tk.CENTER, fill=Tk.BOTH, expand=1 )
        test_frame.pack()

        btn_matrix = []
        for i in range(10):
            btn_row = []
            for j in range(10):
                btn = Tk.Button( test_frame, text='%d,%d' % (i, j), command=lambda i_=i, j_=j: self.test_dialog( j_, i_ ) )
                btn.grid( row=i, column=j, padx=10, pady=10 )
                btn_row.append( btn )
            btn_matrix.append( btn_row )
        self.btn_matrix = btn_matrix

        auto_frame = Tk.Frame( self )
        auto_frame.pack()
        btn = Tk.Button( auto_frame, text='Auto Test', command=self.auto_test )
        btn.pack( pady=20 )

        self.protocol( "WM_DELETE_WINDOW", self.quit )

    def quit( self ):
        self.parent.quit()

    def test_dialog( self, i, j ):
        print( i, j )
        w = self.W_min + int( i * self.W_inc )
        h = self.H_min + int( j * self.H_inc )
        title = '%dx%d' % ( w, h )
        self.dialog = TestDialog( self, title, w, h )
        self.dialog.show()

    def auto_test( self ):
        self.auto = True
        for i in range(10):
            for j in range(10):
                if not self.auto:
                    break
                self.update()
                self.after(  500, lambda: self.dialog.shot() )
                self.after( 1000, lambda: self.dialog.ok() )
                self.btn_matrix[i][j].invoke()
                time.sleep( 1 )

        self.auto = False
        MessageBox.showinfo( "Notify", "Auto Test has been finished.", parent=self )

    def _on_escape( self, *args ):
        self.update()
        yn = MessageBox.askyesno( "Confirmation", "Terminate testing?", parent=self )
        if yn:
            self.auto = False
            # self.dialog.destroy()
            # self.after( 1000, lambda: self.quit() )



# coding: utf-8
"""
    Result.py

    URL: http://stackoverflow.com/questions/1305532/convert-python-dict-to-object

    namedtuple の場合、メンバーの更新は can't set attritube により不可のようなので
    この Struct の方が柔軟。
"""

item_dict = {
    0 : 'Rg',
    1 : 'Rg_stdev',
    2 : 'I0',
    3 : 'I0_stdev',
    4 : 'From',
    5 : 'To',
    6 : 'Quality',
    7 : 'Aggregated',
    8 : 'min_qRg',
    9 : 'max_qRg',
    10 : 'gpfit_I0',
    11 : 'gpfit_Rg',
    12 : 'gpfit_d',
    13 : 'basic_quality',
    14 : 'positive_score',
    15 : 'fit_consistency_pure',
    16 : 'stdev_score',
    17 : 'q_rg_score',
    18 : 'IpI',
    19 : 'bicomponent',
    20 : 'bico_mono_ratio',
    21 : 'bico_G1',
    22 : 'bico_G2',
    23 : 'bico_Rg1',
    24 : 'bico_Rg2',
    25 : 'bico_d1',
    26 : 'bico_d2',
    }

class Result:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

    def __getitem__( self, i ):
        if i < 0:
            i = len( item_dict ) + i
        assert( i >= 0 and i < len( item_dict )  )
        return self.__dict__.get( item_dict[ i ] )

    def get_quality_signal( self ):
        if self.type == 'K':
            color = self.get_kekpf_signal()
        elif self.type == 'E':
            color = None
        else:
            color = self.get_atsas_signal()
        return color

    def get_atsas_signal( self ):
        quality = self.Quality
        aggregated_ = abs( self.Aggregated ) > 0.5

        if quality is None or quality <= 0.3:
            color = 'red'
        elif quality < 0.7 or aggregated_:
            color = 'yellow'
        else:
            color = 'green'

        return color

    def get_kekpf_signal( self ):
        quality = self.Quality

        if quality is None or quality <= 0.3:
            color = 'red'
        elif quality < 0.7:
            color = 'yellow'
        else:
            color = 'green'

        return color




# coding: utf-8
"""
    MicrofluidicMapping.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from scipy.interpolate import UnivariateSpline
from lmfit import minimize, Parameters
from molass_legacy.Mapping.SingleComponent import SingleComponent
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo

DEBUG = False
if DEBUG:
    import molass_legacy.KekLib.DebugPlot as plt

def make_microfluidic_mapping_info_impl(mapper, opt_params):
    mapper.opt_params = opt_params

    mapper.a_base_adjustment  = 0
    mapper.x_base_adjustment  = 0

    compute_mapping_params(mapper)
    determine_mapping_ranges(mapper)

    mapper.a_curve = mapper.make_a_curve(opt_params)
    mapper.x_curve = mapper.make_x_curve(opt_params)
    mapper.x_curve_y_adjusted = mapper.x_curve.y

    if DEBUG:
        fig = plt.figure(figsize=(12,6))
        fig.suptitle("make_microfluidic_mapping_info_impl")
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)
        ax1.plot(mapper.a_curve.y)
        ax1.plot(mapper.a_spline(mapper.a_curve.x))
        ax2.plot(mapper.x_curve_y_adjusted)
        ax2.plot(mapper.x_curve.y)
        ax2.plot(mapper.x_base, color='red')
        fig.tight_layout()
        plt.show()

    mapper.mapped_vector = make_microfluidic_mapped_vector(mapper)
    mapper.mapped_spline  = UnivariateSpline( mapper.x_x, mapper.mapped_vector, s=0, ext=3 )
    mapper.inv_mapped_boundaries = mapper.x_curve.boundaries

    mapper.in_xray_adjustment_mode = False

    mapper.compute_std_diff()

    mapper.scomp = SingleComponent(mapper)
    mapper.logger.info( 'made microfluidic mapping info' )

def compute_mapping_params(mapper):
    mapper.A_init = A_init = len(mapper.a_vector) / len(mapper.x_vector)
    mapper.B_init = B_init = 0
    S_init = mapper.x_vector[-1] / mapper.a_vector[-1]
    params = Parameters()
    params.add( 'A', value=A_init,      min=A_init-0.5, max=A_init+1.0 )
    params.add( 'B', value=B_init,      min=B_init-50,  max=B_init+50 )
    params.add( 'S', value=S_init,      min=S_init*0.25, max=S_init*4 )

    def objective_func(params):
        A = params['A']
        B = params['B']
        S = params['S']
        i = mapper.x_x
        j = A * i + B
        a_y = mapper.a_spline(j)
        return a_y*S - mapper.x_vector

    result = minimize( objective_func, params, args=(), method='least_squares'  )
    A_opt = result.params['A'].value
    B_opt = result.params['B'].value
    S_opt = result.params['S'].value
    mapper.opt_results = [ [ A_opt, B_opt, S_opt ] ]

def determine_mapping_ranges(mapper):
    upper = 0.9
    ratios = [ 0.1, upper, upper ]
    x_ranges = [[ int(len(mapper.x_vector)*r + 0.5)  for r in ratios ]]
    set_setting('range_type', 5)
    """
        Note that
        setting the same value(upper) and setting range_type to 5 here
        ensures the identification of microfluidic peak in later stages.
        See also AnalysisRangeInfo.upgrade_ranges.
    """

    analysis_range_info = AnalysisRangeInfo(x_ranges)
    set_setting('analysis_range_info', analysis_range_info)

    mapper.x_ranges = x_ranges
    mapper.mapping_ranges = x_ranges
    mapper.peak_eval_ranges = x_ranges
    mapper.uv_peak_eval_ranges = [[ int(len(mapper.a_vector)*r + 0.5)  for r in ratios ]]

def make_microfluidic_mapped_vector(mapper):
    opt_result = mapper.opt_results[0]
    A_opt = opt_result[0]
    B_opt = opt_result[1]
    S_opt = opt_result[2]
    return mapper.a_spline( A_opt * mapper.x_x + B_opt ) * S_opt



# coding: utf-8
"""
    QuantileRegressionSklearn.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from sklearn.ensemble import GradientBoostingRegressor

class Model:
    def __init__(self):
        self.model = GradientBoostingRegressor(loss='quantile', alpha=0.5,
                                n_estimators=10, max_depth=3,
                                learning_rate=.1, min_samples_leaf=9,
                                min_samples_split=9)

    def fit(self, x, y):
        X = np.atleast_2d(x).T
        self.model.fit(X, y)

    def predict(self, x):
        X = np.atleast_2d(x).T
        return self.model.predict(X)



"""
    CharFunc/CfDomainStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

def study1():
    from SecTheory.BasicModels import single_pore_pdf
    from SecTheory.SecCF import simple_phi, shifted_phi
    from SecTheory.SecPDF import FftInvImpl

    fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(20,8))
    ax1, ax2, ax3, ax4 = axes[0,:]
    ax5, ax6, ax7, ax8 = axes[1,:]
    ax1.set_title("PDF Domain")
    ax2.set_title("CF Real Domain")
    ax3.set_title("CF Imaginary Domain")
    x = np.arange(21, 321)
    x0 = 20
    npi = 100
    tpi = 1
    y1 = single_pore_pdf(x, npi, tpi)
    y2 = single_pore_pdf(x - x0, npi, tpi)
    # y3 = single_pore_pdf(x - x0 - 10, npi, tpi)
    ty = y1 + y2
    N = 1024
    tz = np.fft.fft(ty, N)
    w = np.fft.fftfreq(N)
    u = np.fft.ifft(tz, len(x))
    z1 = np.fft.fft(y1, N)
    z2 = np.fft.fft(y2, N)
    # z3 = np.fft.fft(y3)

    """
    relation between w and w_
    """

    finv = FftInvImpl()
    w_ = finv.get_w()
    z1_ = simple_phi(w_, npi, tpi)
    u1_ = finv.compute(x, z1_)

    z2_ = shifted_phi(w_, npi, tpi, x0)
    u2_ = finv.compute(x, z2_)

    def objective(p):
        z_ = shifted_phi(w_, npi, tpi, p[0])
        sz = z1_ + z_
        return (np.real(sz) - np.real(tz))**2 + (np.imag(sz) - np.imag(tz))**2

    # res = minimize(objective, [10])
    # print("res.x", res.x)

    ax1.plot(x, ty, label="data")
    ax1.plot(x, y1, ":", label="component-1")
    ax1.plot(x, y2, ":", label="component-2")
    # ax1.plot(x, y3, ":", label="component-3")

    ax1.plot(x, u1_, ":", color="yellow", label="component-1 inverse FFT")
    ax1.plot(x, u2_, ":", color="yellow", label="component-2 inverse FFT")

    tu_ = finv.compute(x, z1_ + z2_)
    ax1.plot(x, tu_, ":", color="red", label="sum inverse FFT")

    print("real(tz)=", np.real(tz))

    ax2.plot(w, np.real(tz), "o", markersize=1, alpha=0.5)
    p2, = ax2.plot(w[0], np.real(tz)[0], "o", color="red", markersize=5, alpha=0.5)
    p2_, = ax2.plot(w[-1], np.real(tz)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax3.plot(w, np.imag(tz), "o", markersize=1, alpha=0.5)
    p3, = ax3.plot(w[0], np.imag(tz)[0], "o", color="red", markersize=5, alpha=0.5)
    p3_, = ax3.plot(w[-1], np.imag(tz)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax2.plot(w, np.real(z1), "o", markersize=1, alpha=0.5)
    ax2.plot(w, np.real(z2), "o", markersize=1, alpha=0.5)
    ax3.plot(w, np.imag(z1), "o", markersize=1, alpha=0.5)
    ax3.plot(w, np.imag(z2), "o", markersize=1, alpha=0.5)

    if False:
        ax1.plot(x, u, ":", color="cyan", label="inverse FFT")
        su = np.fft.ifft(z1 + z2, len(x))
        ax1.plot(x, su, ":", color="red", label="inverse FFT of sum")
    ax1.legend()

    ax4.plot(w, "o", markersize=1, label="w")
    ax4.plot(w_/(2*np.pi), "o", markersize=1, label="w_")
    v_ = np.concatenate([w_[N//2:], w_[:N//2]])
    ax4.plot(v_/(2*np.pi), "o", markersize=1, label="v_")
    ax4.legend()

    tz_ = z1_ + z2_
    print("diff w : v_/(2pi) =", np.max(np.abs(w - v_/(2*np.pi))))
    print("diff tz : tz_ =", np.max(np.abs(np.real(tz) - np.real(tz_))))
    # sz_ = np.concatenate([np.flip(tz_[N//2:]), np.flip(tz_[:N//2])])
    # print("diff tz : sz_ =", np.max(np.abs(np.real(tz) - np.real(sz_))))

    for ax in [ax5, ax8]:
        ax.set_axis_off()

    sz_ = np.concatenate([tz_[N//2:], tz_[:N//2]])
    ax6.plot(v_, np.real(sz_), "o", markersize=1, alpha=0.5)
    p6, = ax6.plot(v_[0], np.real(sz_)[0], "o", color="red", markersize=5, alpha=0.5)
    p6_, = ax6.plot(v_[-1], np.real(sz_)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax7.plot(v_, np.imag(sz_), "o", markersize=1, alpha=0.5)
    p7, = ax7.plot(v_[0], np.imag(sz_)[0], "o", color="red", markersize=5, alpha=0.5)
    p7_, = ax7.plot(v_[-1], np.imag(sz_)[-1], "o", color="blue", markersize=5, alpha=0.5)    
    ax6.plot(w_, np.real(z1_), "o", markersize=1, alpha=0.5)
    ax6.plot(w_, np.real(z2_), "o", markersize=1, alpha=0.5)
    ax7.plot(w_, np.imag(z1_), "o", markersize=1, alpha=0.5)
    ax7.plot(w_, np.imag(z2_), "o", markersize=1, alpha=0.5)

    def on_click(event):
        nonlocal p2, p6, p3, p7
        if event.xdata and event.ydata:
            if event.inaxes == ax2:
                p, q = p2, p6
            elif event.inaxes == ax3:
                p, q = p3, p7
            else:
                return
            
            index = np.argmin((w - event.xdata)**2 + (np.real(tz) - event.ydata)**2)
            print("index=", index)
            px, py = w[index], np.real(tz)[index]
            p.set_data([px], [py])
            px_, py_ = v_[index], np.real(sz_)[index]
            q.set_data([px_], [py_])
            fig.canvas.draw()

    fig.tight_layout()
    fig.canvas.mpl_connect('button_press_event', on_click)
    plt.show()

def study2():
    from SecTheory.BasicModels import single_pore_pdf
    from SecTheory.SecCF import simple_phi, shifted_phi
    from SecTheory.SecPDF import BidirectFft

    fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(20,8))
    fig.suptitle("Proof of Numerical Fourier Transform Inversion of the Sum of Two Monopore CFs", fontsize=20)
    ax1, ax2, ax3, ax4 = axes[0,:]
    ax5, ax6, ax7, ax8 = axes[1,:]
    ax1.set_title("PDF Domain")
    ax2.set_title("CF Real Domain")
    ax3.set_title("CF Imaginary Domain")
    ax4.set_title("Frequency Vector Values")
    x = np.arange(21, 321)
    x0 = 20
    npi = 100
    tpi = 1
    y1 = single_pore_pdf(x, npi, tpi)
    y2 = single_pore_pdf(x - x0, npi, tpi)
    ty = y1 + y2

    bifft = BidirectFft(x)

    tz = bifft.compute_z(ty)
    w = bifft.get_w()
    u = bifft.compute_y(tz)

    z1 = bifft.compute_z(y1)
    z1_ = simple_phi(w, npi, tpi)
    u1_ = bifft.compute_y(z1_)

    z2 = bifft.compute_z(y2)
    z2_ = shifted_phi(w, npi, tpi, x0)
    u2_ = bifft.compute_y(z2_)

    # res = minimize(objective, [10])
    # print("res.x", res.x)

    ax1.plot(x, ty, label="data")
    ax1.plot(x, y1, ":", label="component-1")
    ax1.plot(x, y2, ":", label="component-2")
    # ax1.plot(x, y3, ":", label="component-3")

    ax1.plot(x, u1_, ":", color="yellow", label="component-1 inverse FFT")
    ax1.plot(x, u2_, ":", color="yellow", label="component-2 inverse FFT")

    tu_ = bifft.compute_y(z1_ + z2_)
    ax1.plot(x, tu_, ":", color="red", label="sum inverse FFT")

    print("real(tz)=", np.real(tz))

    ax2.plot(w, np.real(tz), "o", markersize=1, alpha=0.5)
    p2, = ax2.plot(w[0], np.real(tz)[0], "o", color="red", markersize=5, alpha=0.5)
    p2_, = ax2.plot(w[-1], np.real(tz)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax3.plot(w, np.imag(tz), "o", markersize=1, alpha=0.5)
    p3, = ax3.plot(w[0], np.imag(tz)[0], "o", color="red", markersize=5, alpha=0.5)
    p3_, = ax3.plot(w[-1], np.imag(tz)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax2.plot(w, np.real(z1), "o", markersize=1, alpha=0.5)
    ax2.plot(w, np.real(z2), "o", markersize=1, alpha=0.5)
    ax3.plot(w, np.imag(z1), "o", markersize=1, alpha=0.5)
    ax3.plot(w, np.imag(z2), "o", markersize=1, alpha=0.5)
    ax1.legend()

    ax4.plot(w, "o", markersize=1, label="w")
    ax4.legend()

    tz_ = z1_ + z2_
    diff_tz = np.abs(np.real(tz) - np.real(tz_))
    print("real diff tz : tz_ =", np.max(diff_tz))
    ax5.set_title("Difference in CF Real Domain")
    ax5.plot(w, diff_tz)
    diff_tz = np.abs(np.imag(tz) - np.imag(tz_))
    print("imag diff tz : tz_ =", np.max(diff_tz))
    ax8.set_title("Difference in CF Imaginary Domain")
    ax8.plot(w, diff_tz)

    ax6.plot(w, np.real(tz_), "o", markersize=1, alpha=0.5)
    p6, = ax6.plot(w[0], np.real(tz_)[0], "o", color="red", markersize=5, alpha=0.5)
    p6_, = ax6.plot(w[-1], np.real(tz_)[-1], "o", color="blue", markersize=5, alpha=0.5)
    ax7.plot(w, np.imag(tz_), "o", markersize=1, alpha=0.5)
    p7, = ax7.plot(w[0], np.imag(tz_)[0], "o", color="red", markersize=5, alpha=0.5)
    p7_, = ax7.plot(w[-1], np.imag(tz_)[-1], "o", color="blue", markersize=5, alpha=0.5)    
    ax6.plot(w, np.real(z1_), "o", markersize=1, alpha=0.5)
    ax6.plot(w, np.real(z2_), "o", markersize=1, alpha=0.5)
    ax7.plot(w, np.imag(z1_), "o", markersize=1, alpha=0.5)
    ax7.plot(w, np.imag(z2_), "o", markersize=1, alpha=0.5)

    def on_click(event):
        nonlocal p2, p6, p3, p7
        if event.xdata and event.ydata:
            if event.inaxes == ax2:
                p, q = p2, p6
            elif event.inaxes == ax3:
                p, q = p3, p7
            else:
                return
            
            index = np.argmin((w - event.xdata)**2 + (np.real(tz) - event.ydata)**2)
            print("index=", index)
            px, py = w[index], np.real(tz)[index]
            p.set_data([px], [py])
            px_, py_ = w[index], np.real(tz_)[index]
            q.set_data([px_], [py_])
            fig.canvas.draw()

    fig.tight_layout()
    fig.canvas.mpl_connect('button_press_event', on_click)
    plt.show()

if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/../")

    # study1()
    study2()


"""
    Models.Stochastic.OligoporePlotImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.Batch.LiteBatch import LiteBatch
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.Models.Stochastic.OligoporeUtils import plot_oligopore_conformance_impl
from molass_legacy.Models.Stochastic.DatablockUtils import get_block_lrf_src

def plot_results_impl(N, T, pszv, pszp, t0v, tscales, in_folders, block_list, params_list, rg_curves=None):
    num_blocks = len(block_list)

    batch = LiteBatch()

    with plt.Dp(scrollable=True):
        fig, axes = plt.subplots(nrows=num_blocks, ncols=2, figsize=(20, 4*num_blocks))
        i = 0
        if len(tscales) == 0:
            tscales = np.ones(num_blocks)
        if rg_curves is None:
            rg_curves = [None] * num_blocks
        for axes_pair, in_folder, block, t0, tscale, params, rg_curve in zip(axes, in_folders, block_list, t0v, tscales, params_list, rg_curves):
            lrf_src = get_block_lrf_src(batch, in_folder)
            T_ = T*tscale
            t0_ = t0*tscale
            plot_block_state(axes_pair, get_in_folder(in_folder), block, N, T_, t0_, pszv, pszp, lrf_src, params, rg_curve=rg_curve)
            i += 1
            if i > 0:
                # break
                pass

        fig.tight_layout()
        dp = plt.get_dp()
        dp.after(500, lambda: dp.geometry("2000x800"))      # task: do this with auto_resize=True in DebugPlot 
        plt.show()

def rounded_list(fmt, values):
    return ', '.join([fmt % v for v in values])

def plot_block_state(axes_pair, in_folder, block, N, T, t0, pszv, pszp, lrf_src, params, rg_curve=None):
    from importlib import reload
    import RgProcess.RgCurveUtils
    reload(RgProcess.RgCurveUtils)
    from RgProcess.RgCurveUtils import plot_rg_curve
    print("plot_block_state", in_folder)
    ax1, ax2 = axes_pair
    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs()
    peaks = lrf_src.get_peaks()
    egh_moments_list = lrf_src.get_egh_moments_list()
    model = lrf_src.model
    plot_info = Struct(x=lrf_src.xr_x, y=lrf_src.xr_y, model=model, peaks=peaks)

    ax1t = ax1.twinx()
    ax2t = ax2.twinx()

    if rg_curve is not None:
        for ax in [ax1t, ax2t]:
            plot_rg_curve(ax, rg_curve, with_qualities=True, quality_scale=100)

    def split_params_for_plot(p, num_pszv=2):
        """
        note that these params correspond to those returned by estimate_oligopore_distribution_separately
        """
        N, K, t0 = p[0:3]
        pszv = p[3:3+num_pszv]
        pszp = p[3+num_pszv:3+2*num_pszv]   # last 
        return N, K, t0, pszv, pszp

    plot_params = split_params_for_plot(params)
    plot_oligopore_conformance_impl(ax1, ax1t, *plot_params, plot_info,
                                    rg_info=(peak_rgs, peak_trs, True))

    K = N * T
    plot_oligopore_conformance_impl(ax2, ax2t, N, K, t0, pszv, pszp, plot_info,
                                    rg_info=(peak_rgs, peak_trs, True))



"""
    SettingsSerializer.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np  # used in the eval, in cases where the value is like 'np.int64(140)'
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

def serialize(items):
    tuple_ = tuple()
    for name in items:
        tuple_ += (get_setting(name),)
    return str(tuple_).replace(" ", "")

def unserialize(items, serialized_str):
    tuple_ = eval(serialized_str)

    for name, value in zip(items, tuple_):
        set_setting(name, value)

ITEMS = ["poresize_bounds", "t0_upper_bound"]

def serialize_for_optimizer():
    return serialize(ITEMS)

def unserialize_for_optimizer(serialized_str):
    unserialize(ITEMS, serialized_str)



"""
    Models/RateTheory/RobustEDM.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.widgets import Slider
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.Edm import edm_func
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.ElutionModelUtils import compute_4moments, get_xies_from_height_ratio, x_from_height_ratio_impl

VERY_SMALL_VALUE = 1e-8

def edm_impl(x, t0, u, a, b, e, Dz, cinj):
    return edm_func(x-t0, u, a, b, e, Dz, cinj)

def guess_init_params(M):
    M1 = M[1]
    M2 = M[2]
    t0 = M1/2
    u = 30/M1*3
    a = 0.5
    b = -4.0
    e = 0.4
    print("guess_init_params: M2=", M2)
    Dz = 0.2
    cinj = M[0]/2.0 * 0.2
    return np.array([t0, u, a, b, e, Dz, cinj])

def try_optimize(x, y, init_params, debug=False):

    def objective(p):
        y_ = edm_impl(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)
    score = np.sqrt(ret.fun)/np.max(y)

    if debug:
        print("init_params=", init_params)
        print("ret.x=", ret.x)
        print("score=", score)


        def add_sliders(fig, line, slider_axes, sliders, left, width, slider_params):
            slider_specs = [    ("t0", -50, 200, slider_params[0]),
                                ("u", 0, 5, slider_params[1]),
                                ("a", 0, 2, slider_params[2]),
                                ("b", -4, 1, slider_params[3]),
                                ("e", 0, 2, slider_params[4]),
                                ("Dz", 0, 1.5, slider_params[5]),
                                ("cinj", 0, 3, slider_params[6]),
                                # ("tinj", 0, 10, slider_params[7]),
                                ]

            def slider_update(k, val):
                # print([k], "slider_update", val)
                init_params[k] = val
                y_ = edm_impl(x, *init_params)
                line.set_data(x, y_)
                dp = plt.get_dp()
                dp.draw()

            for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                ax_ = fig.add_axes([left, 0.8 - 0.08*k, width, 0.03])
                slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                slider_axes.append(ax_)
                sliders.append(slider)

        with plt.Dp():
            fig, (ax, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax.plot(x, y)

            edm_line1, = ax.plot(x, edm_impl(x, *init_params))
            ax2.plot(x, y)
            edm_line2, = ax2.plot(x, edm_impl(x, *ret.x))

            slider_axes = []
            sliders = []
            add_sliders(fig, edm_line1, slider_axes, sliders, 0.15, 0.25, init_params)
            add_sliders(fig, edm_line2, slider_axes, sliders, 0.65, 0.25, ret.x)

            fig.tight_layout()
            plt.show()

    return ret.x, score

def guess_multiple(x, y, debug=False):
    M = compute_4moments(x, y)
    init_params = guess_init_params(M)
    params, score = try_optimize(x, y, init_params, debug=True)

    y_ = edm_impl(x, *params)

    try_narrower = False

    if try_narrower:
        max_y = np.max(y)
        xL, xR = get_xies_from_height_ratio(0.3, x, y_, max_y=max_y, debug=False)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, y_)
        if try_narrower:
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for px in xL, xR:
                ax.plot([px, px], [ymin, ymax], ":", color="gray")
        fig.tight_layout()
        plt.show()

    if False:

        i, j = [ int(round(px - x[0])) for px in (xL, xR) ]
        slice_ = slice(i,j+1)
        sx = x[slice_]
        sy = y[slice_]
        better_params, score = try_optimize(sx, sy, params)

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y)
            ax.plot(x, y_)
            ax.plot(x, edm_impl(x, *better_params))
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for px in xL, xR:
                ax.plot([px, px], [ymin, ymax], ":", color="gray")
            fig.tight_layout()
            plt.show()

class EDM(Model):
    def __init__(self, **kwargs):
        super(EDM, self).__init__(edm_impl, **kwargs)

    def get_name(self):
        return "EDM"

    def is_traditional(self):
        return False

    def eval(self, params=None, x=None):
        return self.func(x, *params)

    def x_from_height_ratio(self, ecurve, ratio, params):
        return x_from_height_ratio_impl(edm_impl, ecurve, ratio, *params, needs_ymax=True, full_params=True)

    def get_params_string(self, params):
        return 't0=%g, u=%g, a=%g, b=%g, e=%g, Dz=%g, cinj=%g' % tuple(params)



"""
    Global.V2Init.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
from molass_legacy.Experiment.DataUtils import get_columntype, update_sec_settings_impl

def update_sec_settings(el_option=0, column_name=None, exclusion_limit=None, num_plates_pm=None):

    if column_name is None:
        columntype = get_columntype()
        column_name = columntype.name
        if num_plates_pm is None:
            num_plates_pm = columntype.num_plates

    update_sec_settings_impl(el_option, column_name, exclusion_limit, num_plates_pm)



"""
    Optimizer.TheUtils.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting

FILES = [
    "callback.txt",
    "in_data_info.txt",
    "init_params.txt",
    "optimizer.log",
    "pid.txt",
    "seed.txt",
    "trimming.txt",
    "bounds.txt",
    ]

def get_optimizer_folder():
    optimizer_folder = get_setting('optimizer_folder')
    if optimizer_folder is None:
        optimizer_folder = os.path.join(get_setting('analysis_folder'), 'optimized')
    return optimizer_folder

def get_treatment_path(optimizer_folder=None):
    from molass_legacy.Optimizer.TheUtils import get_optimizer_folder
    if optimizer_folder is None:
        optimizer_folder = get_optimizer_folder()
    return os.path.join(optimizer_folder, "treatment.txt")

def get_optjob_folder_impl():
    return os.path.join(get_optimizer_folder(), 'jobs').replace('\\', '/')

def get_analysis_folder_from_work_folder(work_folder):
    # analysis_folder <= analysis_folder/optimized/jobs/nnn
    return os.path.abspath(work_folder + '/../../..')

def guess_n_compnents_trial(dsets, logger):

    (xr_curve, D), rg_curve, (uv_curve, U) = dsets

    epeaks = xr_curve.get_emg_peaks()
    n = len(epeaks)
    n_compnents = n + 2

    logger.info("n_components guessed as %d from len(epeaks)=%d", n_compnents, n)

    if False:
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.Elution.CurveUtils import simple_plot

        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
        simple_plot(ax1, uv_curve, color='blue')
        simple_plot(ax2, xr_curve, color='orange')
        fig.tight_layout()
        plt.show()
        plt.pop()

    return n_compnents

def get_work_info(work_folder):
    import os
    import re
    from molass_legacy.KekLib.BasicUtils import Struct
    in_data_txt = os.path.join(work_folder, FILES[1])
    in_folder_re = re.compile(r"in_folder=(\S+)")
    in_folder = None
    with open(in_data_txt) as fh:
        for line in fh:
            m = in_folder_re.search(line)
            if m:
                in_folder = m.group(1)
                break

    init_params_txt = os.path.join(work_folder, FILES[2])
    init_params = np.loadtxt(init_params_txt)

    bounds_txt = os.path.join(work_folder, FILES[7])
    if os.path.exists(bounds_txt):
        real_bounds = np.loadtxt(bounds_txt)
    else:
        real_bounds = None

    return  Struct(work_folder=work_folder, in_folder=in_folder, init_params=init_params, real_bounds=real_bounds)

def get_work_info_for_demo():
    from molass_legacy._MOLASS.SerialSettings import get_setting

    work_folder = get_setting("optjob_folder")
    if work_folder is None:
        in_folder = get_setting('in_folder')
        if in_folder.find('20181203') > 0:
            work_folder = r'D:\TODO\optimizer\result-20210624\20181203\000'
        else:
            work_folder = r'D:\TODO\20210628\optimizer-results\temp-save-32\optimizer\000'

    return get_work_info(work_folder)

def convert_to_peak_significances(xr_curve, xr_params_list):
    score_list = []
    for h, mu, sigma, tau in xr_params_list:
        score_list.append(h*sigma)
    score_array = np.array(score_list)
    return score_array/np.sum(score_array)

def get_sd_from_folder_impl(in_folder, logger, ver_date=None):
    from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.SecSaxs.DataTreatment import DataTreatment

    logger.info("loading data from %s using StandardProcedure", in_folder)
    sp = StandardProcedure()
    sd_raw = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd_raw)

    treat = DataTreatment()
    treat.load()        # this requires "optimizer_folder" setting
    if ver_date is None or ver_date >= "2023-01-01":
        sd = treat.get_trimmed_sd(sd_raw, pre_recog)
        corrected_sd = treat.get_corrected_sd(sd_raw, pre_recog, sd)
    else:
        sd = treat.get_treated_sd(sd_raw, pre_recog)
        corrected_sd = sd
    sd.conc_factor = compute_conc_factor_util()
    return sd, corrected_sd, treat



# coding: utf-8
"""

    Test.TestUtils.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF

"""
from molass_legacy.SerialAnalyzer.DataUtils import serial_folder_walk

def execute_for_all_data(root_folder, func):
    serial_folder_walk(root_folder, func)

def show_test_dialog(root, in_folder, prepare_func, tester_func, frame_func=None, v2_standard=True):
    from molass_legacy.KekLib.OurTkinter import Tk, Dialog
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from SecSaxs.DataTreatment import DataTreatment
    from molass_legacy.Optimizer.StateSequence import StateSequence
    from molass_legacy.Optimizer.OptimizerSettings import OptimizerSettings
    from molass_legacy.Optimizer.InitialInfo import InitialInfo
    from molass_legacy.Optimizer.OptDataSets import OptDataSets
    from molass_legacy.Optimizer.OptJobInfo import OptJobInfo
    from molass_legacy.KekLib.BasicUtils import Struct

    if root is None:
        from molass_legacy.KekLib.TkUtils import get_tk_root
        root = get_tk_root()

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    if v2_standard:
        pre_recog = PreliminaryRecognition(sd)
        treat = DataTreatment(route="v2", trimming=2, correction=1)
        sd_copy = treat.get_treated_sd(sd, pre_recog)
    else:
        pre_recog = None
        treat = None
        sd_copy = sd

    pref_info = prepare_func(treat, sd_copy, pre_recog)

    class TestDialog(Dialog):
        def __init__(self, parent):
            Dialog.__init__(self, parent, "TestDialog", visible=False)

        def show(self):
            self._show()

        def body(self, body_frame):
            if frame_func is not None:
                frame_func(self, body_frame)

            run_btn = Tk.Button(body_frame, text="Run", command=self.run)
            run_btn.pack()

        def run(self):
            tester_func(self, pref_info)

    def show_pi():
        dialog = TestDialog(root)
        dialog.show()
        root.quit()

    root.after(100, show_pi)
    root.mainloop()
    root.destroy()



"""
    MatchingPeaks.py

    Copyright (c) 2020-2022, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from itertools import combinations

ALLOWANCE_RATIO = 0.05

def compute_matching_indeces(size, uv_top_x, xr_top_x, logger=None):
    """
    e.g.,
                   input           output (return)
        uv_top_x = [189, 625]   => [   0, 1, None]
        xr_top_x = [624, 800]   => [None, 0,    1]
    """
    ret_list1 = []
    ret_list2 = []

    if len(uv_top_x) >= len(xr_top_x):
        list_more = uv_top_x
        list_less = xr_top_x
        reversed_ = False
    else:
        list_more = xr_top_x
        list_less = uv_top_x
        reversed_ = True

    less_indeces = list(range(len(list_less)))
    greater_indeces = list(range(len(list_more)))

    ret_pairs = []
    start_j = 0
    matched = set()
    unmatched_peaks = []
    for i in greater_indeces:
        found = False
        for j in less_indeces[start_j:]:
            dist = abs(list_more[i] - list_less[j])
            if dist/size < ALLOWANCE_RATIO:
                ret_pairs.append((i,j))
                start_j = j + 1
                found = True
                matched.add(j)
                break
        if not found:
            ret_pairs.append((i,None))
            unmatched_peaks.append(list_more[i])

    if len(matched) < len(list_less):
        unmatched = list(set(less_indeces) - matched)
        for k in unmatched:
            peak = list_less[k]
            i = bisect_right(list_more, peak)
            ret_pairs.insert(i, (None, k))
            unmatched_peaks.append(list_less[k])

    if len(unmatched_peaks) > 0 and logger is not None:
        logger.warning("compute_matching_indeces detected unmatched peaks in : %s", str(sorted(unmatched_peaks)))

    list1, list2 = list(zip(*ret_pairs))
    if reversed_:
        list2, list1 = list1, list2
    return list1, list2

class MatchingPeaks:
    def __init__(self, cs, uv_curve_, xr_curve_, debug=False):
        self.logger = logging.getLogger(__name__)

        self.uv_curve = uv_curve = cs.a_curve
        self.xr_curve = xr_curve = cs.x_curve

        A = cs.slope
        B = cs.intercept

        uv_peaks = uv_curve.get_emg_peaks()
        xr_peaks = xr_curve.get_emg_peaks()
        uv_top_x = [int(p.top_x) for p in uv_peaks]
        xr_top_x = [int(p.top_x*A + B)  for p in xr_peaks]
        ii, jj = compute_matching_indeces(len(uv_curve.y), uv_top_x, xr_top_x, self.logger)
        print("ii=", ii)
        print("jj=", jj)

        ret_uv_peaks = []
        ret_xr_peaks = []
        for i, j in zip(ii, jj):
            if i is None or j is None:
                # raise PeakMathcingError()
                """
                seems to be degraded after compute_matching_indeces changed
                for Mapping.FewerPointMapper.py, which is not used yet.
                """
                continue
            ret_uv_peaks.append(uv_peaks[i])
            ret_xr_peaks.append(xr_peaks[j])

        if debug:
            self.debug_plot_emg_peaks("get_matching_peak_lists", xr_peaks, ret_xr_peaks, uv_peaks, ret_uv_peaks)

        self.ii = ii
        self.jj = jj
        self.uv_peaks = ret_uv_peaks
        self.xr_peaks = ret_xr_peaks

    def debug_plot_emg_peaks(self, title, xr_emg_peaks, ret_xr_peaks, uv_emg_peaks, ret_uv_peaks):
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.Elution.CurveUtils import simple_plot

        plt.push()
        xr_ecurve = self.xr_curve
        uv_ecurve = self.uv_curve

        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(14,10))

        fig.suptitle(title)

        def emg_peak_plot(ax, ecurve, color, emg_peaks):
            simple_plot(ax, ecurve, color=color, legend=False)
            x = ecurve.x
            for k, epeak in enumerate(emg_peaks):
                y = epeak.get_model_y(x)
                ax.plot(x, y, ':', label=str(k))
            ax.legend()

        emg_peak_plot(axes[0,0], xr_ecurve, "orange", xr_emg_peaks)
        emg_peak_plot(axes[0,1], xr_ecurve, "orange", ret_xr_peaks)
        x = uv_ecurve.x
        emg_peak_plot(axes[1,0], uv_ecurve, "blue", uv_emg_peaks)
        emg_peak_plot(axes[1,1],  uv_ecurve, "blue", ret_uv_peaks)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        plt.show()
        plt.pop()

    def get_matching_peak_lists(self):
        return self.uv_peaks, self.xr_peaks



import numpy    as np
import matplotlib.pyplot    as plt
import seaborn              as sns
from Intensity              import f

x = np.linspace( 0, 10, 100 )

plt.title( 'f(sR) = 3 * ( np.sin( sR ) - sR * np.cos( sR ) ) / ( sR )**3' )
plt.ylabel( 'Intensity' )
plt.xlabel( 'sR' )
plt.plot( f(x) )
plt.show()
